# distutils: language=c++
# cython: language_level=3, binding=True, linetrace=True

from rapidfuzz.fuzz import WRatio, ratio

cimport cython
from cpp_common cimport (
    PyObjectWrapper,
    RF_KwargsWrapper,
    RF_StringWrapper,
    conv_sequence_with_none,
    conv_sequence,
    get_score_cutoff_f64,
    get_score_cutoff_i64,
    get_score_cutoff_size_t,
    is_none
)
from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_F_CONTIGUOUS, PyBUF_ND, PyBUF_SIMPLE
from cpython.exc cimport PyErr_CheckSignals
from cpython.list cimport PyList_New, PyList_SET_ITEM
from cpython.object cimport PyObject
from cpython.ref cimport Py_INCREF
from libcpp.cmath cimport floor, isnan
from libc.stdint cimport int32_t, int64_t, uint8_t, uint64_t
from libcpp cimport algorithm, bool
from libcpp.utility cimport move
from libcpp.vector cimport vector

import heapq
from array import array
import sys

from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_IsValid

from rapidfuzz cimport (
    RF_SCORER_FLAG_RESULT_F64,
    RF_SCORER_FLAG_RESULT_I64,
    RF_SCORER_FLAG_RESULT_SIZE_T,
    RF_SCORER_FLAG_SYMMETRIC,
    RF_SCORER_NONE_IS_WORST_SCORE,
    SCORER_STRUCT_VERSION,
    PREPROCESSOR_STRUCT_VERSION,
    RF_Kwargs,
    RF_Preprocess,
    RF_Preprocessor,
    RF_Scorer,
    RF_ScorerFlags,
    RF_ScorerFunc,
    RF_String,
)

pandas_NA = None

cdef inline void setupPandas() noexcept:
    global pandas_NA
    if pandas_NA is None:
        pandas = sys.modules.get('pandas')
        if hasattr(pandas, 'NA'):
            pandas_NA = pandas.NA

setupPandas()

cdef extern from "process_cpp.hpp":
    cdef cppclass ExtractComp:
        ExtractComp()
        ExtractComp(const RF_ScorerFlags* scorer_flags)

    cdef cppclass ListMatchElem[T]:
        T score
        int64_t index
        PyObjectWrapper choice

    cdef cppclass DictMatchElem[T]:
        T score
        int64_t index
        PyObjectWrapper choice
        PyObjectWrapper key

    cdef cppclass DictStringElem:
        DictStringElem()
        DictStringElem(int64_t index, PyObjectWrapper key, PyObjectWrapper val, RF_StringWrapper proc_val)

        int64_t index
        PyObjectWrapper key
        PyObjectWrapper val
        RF_StringWrapper proc_val

    cdef cppclass ListStringElem:
        ListStringElem()
        ListStringElem(int64_t index, PyObjectWrapper val, RF_StringWrapper proc_val)

        int64_t index
        PyObjectWrapper val
        RF_StringWrapper proc_val

    cdef cppclass RF_ScorerWrapper:
        RF_ScorerFunc scorer_func

        RF_ScorerWrapper()
        RF_ScorerWrapper(RF_ScorerFunc)

        void call(const RF_String*, double, double, double*) except +
        void call(const RF_String*, int64_t, int64_t, int64_t*) except +
        void call(const RF_String*, size_t, size_t, size_t*) except +

    cdef vector[DictMatchElem[T]] extract_dict_impl[T](
        const RF_Kwargs*, const RF_ScorerFlags*, RF_Scorer*,
        const RF_StringWrapper&, const vector[DictStringElem]&, T, T) except +

    cdef vector[ListMatchElem[T]] extract_list_impl[T](
        const RF_Kwargs*, const RF_ScorerFlags*, RF_Scorer*,
        const RF_StringWrapper&, const vector[ListStringElem]&, T, T) except +

    cdef bool is_lowest_score_worst[T](const RF_ScorerFlags* scorer_flags)
    cdef T get_optimal_score[T](const RF_ScorerFlags* scorer_flags)

    cpdef enum class MatrixType:
        UNDEFINED = 0
        FLOAT32 = 1
        FLOAT64 = 2
        INT8 = 3
        INT16 = 4
        INT32 = 5
        INT64 = 6
        UINT8 = 7
        UINT16 = 8
        UINT32 = 9
        UINT64 = 10

    cdef cppclass RfMatrix "Matrix":
        RfMatrix() except +
        RfMatrix(MatrixType, size_t, size_t) except +
        int get_dtype_size() except +
        const char* get_format() except +
        void set[T](size_t, size_t, T) except +

        MatrixType m_dtype
        size_t m_rows
        size_t m_cols
        void* m_matrix
        bool vector_output

    RfMatrix cdist_single_list_impl[T](  const RF_ScorerFlags* scorer_flags, const RF_Kwargs*, RF_Scorer*,
        const vector[RF_StringWrapper]&, MatrixType, int, T, T, T) except +
    RfMatrix cdist_two_lists_impl[T](    const RF_ScorerFlags* scorer_flags, const RF_Kwargs*, RF_Scorer*,
        const vector[RF_StringWrapper]&, const vector[RF_StringWrapper]&, MatrixType, int, T, T, T) except +
    RfMatrix cpdist_cpp_impl[T](    const RF_Kwargs*, RF_Scorer*,
        const vector[RF_StringWrapper]&, const vector[RF_StringWrapper]&, MatrixType, int, T, T, T) except +


cdef inline vector[DictStringElem] preprocess_dict(queries, processor) except *:
    cdef vector[DictStringElem] proc_queries
    cdef int64_t queries_len = <int64_t>len(queries)
    cdef RF_String proc_str
    cdef RF_Preprocessor* processor_context = NULL
    proc_queries.reserve(queries_len)
    cdef int64_t i

    # No processor
    if not processor:
        for i, (query_key, query) in enumerate(queries.items()):
            conv_sequence_with_none(query, &proc_str)
            if proc_str.data == NULL:
                continue

            proc_queries.emplace_back(
                i,
                move(PyObjectWrapper(query_key)),
                move(PyObjectWrapper(query)),
                move(RF_StringWrapper(proc_str))
            )
    else:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

        # use RapidFuzz C-Api
        if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            for i, (query_key, query) in enumerate(queries.items()):
                if is_none(query):
                    continue
                processor_context.preprocess(query, &proc_str)
                proc_queries.emplace_back(
                    i,
                    move(PyObjectWrapper(query_key)),
                    move(PyObjectWrapper(query)),
                    move(RF_StringWrapper(proc_str))
                )

        # Call Processor through Python
        else:
            for i, (query_key, query) in enumerate(queries.items()):
                if is_none(query):
                    continue
                proc_query = processor(query)
                proc_queries.emplace_back(
                    i,
                    move(PyObjectWrapper(query_key)),
                    move(PyObjectWrapper(query)),
                    move(RF_StringWrapper(conv_sequence(proc_query), proc_query))
                )

    return move(proc_queries)


cdef inline vector[ListStringElem] preprocess_list(queries, processor) except *:
    cdef vector[ListStringElem] proc_queries
    cdef int64_t queries_len = <int64_t>len(queries)
    cdef RF_String proc_str
    cdef RF_Preprocessor* processor_context = NULL
    proc_queries.reserve(queries_len)
    cdef int64_t i

    # No processor
    if not processor:
        for i, query in enumerate(queries):
            conv_sequence_with_none(query, &proc_str)
            if proc_str.data == NULL:
                continue

            proc_queries.emplace_back(
                i,
                move(PyObjectWrapper(query)),
                move(RF_StringWrapper(proc_str))
            )
    else:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

        # use RapidFuzz C-Api
        if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            for i, query in enumerate(queries):
                if is_none(query):
                    continue
                processor_context.preprocess(query, &proc_str)
                proc_queries.emplace_back(
                    i,
                    move(PyObjectWrapper(query)),
                    move(RF_StringWrapper(proc_str))
                )

        # Call Processor through Python
        else:
            for i, query in enumerate(queries):
                if is_none(query):
                    continue
                proc_query = processor(query)
                proc_queries.emplace_back(
                    i,
                    move(PyObjectWrapper(query)),
                    move(RF_StringWrapper(conv_sequence(proc_query), proc_query))
                )

    return move(proc_queries)


cdef inline extractOne_dict_f64(
    query, choices, RF_Scorer* scorer, const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    cdef RF_String proc_str
    cdef double score
    cdef Py_ssize_t i = 0
    cdef RF_Preprocessor* processor_context = NULL
    if processor:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

    cdef RF_StringWrapper proc_query = move(RF_StringWrapper(conv_sequence(query)))
    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
    cdef double c_score_hint = get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)

    cdef RF_ScorerFunc scorer_func
    scorer.scorer_func_init(&scorer_func, scorer_kwargs, 1, &proc_query.string)
    cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)

    cdef bool lowest_score_worst = is_lowest_score_worst[double](scorer_flags)
    cdef double optimal_score = get_optimal_score[double](scorer_flags)

    cdef bool result_found = False
    cdef double result_score = 0
    result_key = None
    result_choice = None

    for choice_key, choice in choices.items():
        if i % 1000 == 0:
            PyErr_CheckSignals()
        i += 1
        if is_none(choice):
            continue

        if processor is None:
            proc_choice = move(RF_StringWrapper(conv_sequence(choice)))
        elif processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            processor_context.preprocess(choice, &proc_str)
            proc_choice = move(RF_StringWrapper(proc_str))
        else:
            py_proc_choice = processor(choice)
            proc_choice = move(RF_StringWrapper(conv_sequence(py_proc_choice)))

        ScorerFunc.call(&proc_choice.string, c_score_cutoff, c_score_hint, &score)

        if lowest_score_worst:
            if score >= c_score_cutoff and (not result_found or score > result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_key = choice_key
                result_found = True
        else:
            if score <= c_score_cutoff and (not result_found or score < result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_key = choice_key
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_key) if result_found else None


cdef inline extractOne_dict_i64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    cdef RF_String proc_str
    cdef int64_t score
    cdef Py_ssize_t i = 0
    cdef RF_Preprocessor* processor_context = NULL
    if processor:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

    cdef RF_StringWrapper proc_query = move(RF_StringWrapper(conv_sequence(query)))
    cdef int64_t c_score_cutoff = get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
    cdef int64_t c_score_hint = get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)

    cdef RF_ScorerFunc scorer_func
    scorer.scorer_func_init(&scorer_func, scorer_kwargs, 1, &proc_query.string)
    cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)

    cdef bool lowest_score_worst = is_lowest_score_worst[int64_t](scorer_flags)
    cdef int64_t optimal_score = get_optimal_score[int64_t](scorer_flags)

    cdef bool result_found = False
    cdef int64_t result_score = 0
    result_key = None
    result_choice = None

    for choice_key, choice in choices.items():
        if i % 1000 == 0:
            PyErr_CheckSignals()
        i += 1
        if is_none(choice):
            continue

        if processor is None:
            proc_choice = move(RF_StringWrapper(conv_sequence(choice)))
        elif processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            processor_context.preprocess(choice, &proc_str)
            proc_choice = move(RF_StringWrapper(proc_str))
        else:
            py_proc_choice = processor(choice)
            proc_choice = move(RF_StringWrapper(conv_sequence(py_proc_choice)))

        ScorerFunc.call(&proc_choice.string, c_score_cutoff, c_score_hint, &score)

        if lowest_score_worst:
            if score >= c_score_cutoff and (not result_found or score > result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_key = choice_key
                result_found = True
        else:
            if score <= c_score_cutoff and (not result_found or score < result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_key = choice_key
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_key) if result_found else None


cdef inline extractOne_dict_size_t(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    cdef RF_String proc_str
    cdef size_t score
    cdef Py_ssize_t i = 0
    cdef RF_Preprocessor* processor_context = NULL
    if processor:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

    cdef RF_StringWrapper proc_query = move(RF_StringWrapper(conv_sequence(query)))
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
    cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)

    cdef RF_ScorerFunc scorer_func
    scorer.scorer_func_init(&scorer_func, scorer_kwargs, 1, &proc_query.string)
    cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)

    cdef bool lowest_score_worst = is_lowest_score_worst[size_t](scorer_flags)
    cdef size_t optimal_score = get_optimal_score[size_t](scorer_flags)

    cdef bool result_found = False
    cdef size_t result_score = 0
    result_key = None
    result_choice = None

    for choice_key, choice in choices.items():
        if i % 1000 == 0:
            PyErr_CheckSignals()
        i += 1
        if is_none(choice):
            continue

        if processor is None:
            proc_choice = move(RF_StringWrapper(conv_sequence(choice)))
        elif processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            processor_context.preprocess(choice, &proc_str)
            proc_choice = move(RF_StringWrapper(proc_str))
        else:
            py_proc_choice = processor(choice)
            proc_choice = move(RF_StringWrapper(conv_sequence(py_proc_choice)))

        ScorerFunc.call(&proc_choice.string, c_score_cutoff, c_score_hint, &score)

        if lowest_score_worst:
            if score >= c_score_cutoff and (not result_found or score > result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_key = choice_key
                result_found = True
        else:
            if score <= c_score_cutoff and (not result_found or score < result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_key = choice_key
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_key) if result_found else None


cdef inline extractOne_dict(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    flags = scorer_flags.flags

    if flags & RF_SCORER_FLAG_RESULT_F64:
        return extractOne_dict_f64(
            query, choices, scorer, scorer_flags, processor, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        return extractOne_dict_size_t(
            query, choices, scorer, scorer_flags, processor, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        return extractOne_dict_i64(
            query, choices, scorer, scorer_flags, processor, score_cutoff, score_hint, scorer_kwargs
        )

    raise ValueError("scorer does not properly use the C-API")

cdef inline extractOne_list_f64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    cdef RF_String proc_str
    cdef double score
    cdef Py_ssize_t i
    cdef RF_Preprocessor* processor_context = NULL
    if processor:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

    cdef RF_StringWrapper proc_query = move(RF_StringWrapper(conv_sequence(query)))
    cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
    cdef double c_score_hint = get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)

    cdef RF_ScorerFunc scorer_func
    scorer.scorer_func_init(&scorer_func, scorer_kwargs, 1, &proc_query.string)
    cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)

    cdef bool lowest_score_worst = is_lowest_score_worst[double](scorer_flags)
    cdef double optimal_score = get_optimal_score[double](scorer_flags)

    cdef bool result_found = False
    cdef double result_score = 0
    cdef Py_ssize_t result_index = 0
    result_choice = None

    for i, choice in enumerate(choices):
        if i % 1000 == 0:
            PyErr_CheckSignals()
        if is_none(choice):
            continue

        if processor is None:
            proc_choice = move(RF_StringWrapper(conv_sequence(choice)))
        elif processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            processor_context.preprocess(choice, &proc_str)
            proc_choice = move(RF_StringWrapper(proc_str))
        else:
            py_proc_choice = processor(choice)
            proc_choice = move(RF_StringWrapper(conv_sequence(py_proc_choice)))

        ScorerFunc.call(&proc_choice.string, c_score_cutoff, c_score_hint, &score)

        if lowest_score_worst:
            if score >= c_score_cutoff and (not result_found or score > result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_index = i
                result_found = True
        else:
            if score <= c_score_cutoff and (not result_found or score < result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_index = i
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_index) if result_found else None

cdef inline extractOne_list_i64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    cdef RF_String proc_str
    cdef int64_t score
    cdef Py_ssize_t i
    cdef RF_Preprocessor* processor_context = NULL
    if processor:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

    cdef RF_StringWrapper proc_query = move(RF_StringWrapper(conv_sequence(query)))
    cdef int64_t c_score_cutoff = get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
    cdef int64_t c_score_hint = get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)

    cdef RF_ScorerFunc scorer_func
    scorer.scorer_func_init(&scorer_func, scorer_kwargs, 1, &proc_query.string)
    cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)

    cdef bool lowest_score_worst = is_lowest_score_worst[int64_t](scorer_flags)
    cdef int64_t optimal_score = get_optimal_score[int64_t](scorer_flags)

    cdef bool result_found = False
    cdef int64_t result_score = 0
    cdef Py_ssize_t result_index = 0
    result_choice = None

    for i, choice in enumerate(choices):
        if i % 1000 == 0:
            PyErr_CheckSignals()
        if is_none(choice):
            continue

        if processor is None:
            proc_choice = move(RF_StringWrapper(conv_sequence(choice)))
        elif processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            processor_context.preprocess(choice, &proc_str)
            proc_choice = move(RF_StringWrapper(proc_str))
        else:
            py_proc_choice = processor(choice)
            proc_choice = move(RF_StringWrapper(conv_sequence(py_proc_choice)))

        ScorerFunc.call(&proc_choice.string, c_score_cutoff, c_score_hint, &score)

        if lowest_score_worst:
            if score >= c_score_cutoff and (not result_found or score > result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_index = i
                result_found = True
        else:
            if score <= c_score_cutoff and (not result_found or score < result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_index = i
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_index) if result_found else None

cdef inline extractOne_list_size_t(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    cdef RF_String proc_str
    cdef size_t score
    cdef Py_ssize_t i
    cdef RF_Preprocessor* processor_context = NULL
    if processor:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

    cdef RF_StringWrapper proc_query = move(RF_StringWrapper(conv_sequence(query)))
    cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
    cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)

    cdef RF_ScorerFunc scorer_func
    scorer.scorer_func_init(&scorer_func, scorer_kwargs, 1, &proc_query.string)
    cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)

    cdef bool lowest_score_worst = is_lowest_score_worst[size_t](scorer_flags)
    cdef size_t optimal_score = get_optimal_score[size_t](scorer_flags)

    cdef bool result_found = False
    cdef size_t result_score = 0
    cdef Py_ssize_t result_index = 0
    result_choice = None

    for i, choice in enumerate(choices):
        if i % 1000 == 0:
            PyErr_CheckSignals()
        if is_none(choice):
            continue

        if processor is None:
            proc_choice = move(RF_StringWrapper(conv_sequence(choice)))
        elif processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            processor_context.preprocess(choice, &proc_str)
            proc_choice = move(RF_StringWrapper(proc_str))
        else:
            py_proc_choice = processor(choice)
            proc_choice = move(RF_StringWrapper(conv_sequence(py_proc_choice)))

        ScorerFunc.call(&proc_choice.string, c_score_cutoff, c_score_hint, &score)

        if lowest_score_worst:
            if score >= c_score_cutoff and (not result_found or score > result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_index = i
                result_found = True
        else:
            if score <= c_score_cutoff and (not result_found or score < result_score):
                result_score = c_score_cutoff = score
                result_choice = choice
                result_index = i
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_index) if result_found else None

cdef inline extractOne_list(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    flags = scorer_flags.flags

    if flags & RF_SCORER_FLAG_RESULT_F64:
        return extractOne_list_f64(
            query, choices, scorer, scorer_flags, processor, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        return extractOne_list_size_t(
            query, choices, scorer, scorer_flags, processor, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        return extractOne_list_i64(
            query, choices, scorer, scorer_flags, processor, score_cutoff, score_hint, scorer_kwargs
        )

    raise ValueError("scorer does not properly use the C-API")


cdef inline get_scorer_flags_py(scorer, dict scorer_kwargs):
    params = getattr(scorer, '_RF_ScorerPy', None)
    if params is not None:
        flags = params["get_scorer_flags"](**scorer_kwargs)
        return (flags["worst_score"], flags["optimal_score"])
    return (0, 100)

cdef inline py_extractOne_dict(query, choices, scorer, processor, double score_cutoff, worst_score, optimal_score, dict scorer_kwargs):
    cdef bool lowest_score_worst = optimal_score > worst_score
    cdef bool result_found = False
    result_score = 0
    result_choice = None
    result_key = None

    for choice_key, choice in choices.items():
        if is_none(choice):
            continue

        if processor is not None:
            score = scorer(query, processor(choice), **scorer_kwargs)
        else:
            score = scorer(query, choice, **scorer_kwargs)

        if lowest_score_worst:
            if score >= score_cutoff and (not result_found or score > result_score):
                score_cutoff = score
                scorer_kwargs["score_cutoff"] = score
                result_score = score
                result_choice = choice
                result_key = choice_key
                result_found = True
        else:
            if score <= score_cutoff and (not result_found or score < result_score):
                score_cutoff = score
                scorer_kwargs["score_cutoff"] = score
                result_score = score
                result_choice = choice
                result_key = choice_key
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_key) if result_choice is not None else None


cdef inline py_extractOne_list(query, choices, scorer, processor, double score_cutoff, worst_score, optimal_score, dict scorer_kwargs):
    cdef bool lowest_score_worst = optimal_score > worst_score
    cdef bool result_found = False
    cdef int64_t result_index = 0
    cdef int64_t i
    result_score = 0
    result_choice = None

    for i, choice in enumerate(choices):
        if is_none(choice):
            continue

        if processor is not None:
            score = scorer(query, processor(choice), **scorer_kwargs)
        else:
            score = scorer(query, choice, **scorer_kwargs)

        if lowest_score_worst:
            if score >= score_cutoff and (not result_found or score > result_score):
                score_cutoff = score
                scorer_kwargs["score_cutoff"] = score
                result_score = score
                result_choice = choice
                result_index = i
                result_found = True
        else:
            if score <= score_cutoff and (not result_found or score < result_score):
                score_cutoff = score
                scorer_kwargs["score_cutoff"] = score
                result_score = score
                result_choice = choice
                result_index = i
                result_found = True

        if score == optimal_score:
            break

    return (result_choice, result_score, result_index) if result_choice is not None else None


def extractOne(query, choices, *, scorer=WRatio, processor=None, score_cutoff=None, score_hint=None, scorer_kwargs=None):
    cdef RF_Scorer* scorer_context = NULL
    cdef RF_ScorerFlags scorer_flags

    scorer_kwargs = scorer_kwargs.copy() if scorer_kwargs else {}

    setupPandas()

    if is_none(query):
        return None

    # preprocess the query
    if callable(processor):
        query = processor(query)


    scorer_capsule = getattr(scorer, '_RF_Scorer', scorer)
    if PyCapsule_IsValid(scorer_capsule, NULL):
        scorer_context = <RF_Scorer*>PyCapsule_GetPointer(scorer_capsule, NULL)

    if scorer_context and scorer_context.version == SCORER_STRUCT_VERSION:
        kwargs_context = RF_KwargsWrapper()
        scorer_context.kwargs_init(&kwargs_context.kwargs, scorer_kwargs)
        scorer_context.get_scorer_flags(&kwargs_context.kwargs, &scorer_flags)

        if hasattr(choices, "items"):
            return extractOne_dict(query, choices, scorer_context, &scorer_flags,
                processor, score_cutoff, score_hint, &kwargs_context.kwargs)
        else:
            return extractOne_list(query, choices, scorer_context, &scorer_flags,
                processor, score_cutoff, score_hint, &kwargs_context.kwargs)


    worst_score, optimal_score = get_scorer_flags_py(scorer, scorer_kwargs)
    # the scorer has to be called through Python
    if score_cutoff is None:
        score_cutoff = worst_score

    scorer_kwargs["score_cutoff"] = score_cutoff

    if hasattr(choices, "items"):
        return py_extractOne_dict(query, choices, scorer, processor, score_cutoff, worst_score, optimal_score, scorer_kwargs)
    else:
        return py_extractOne_list(query, choices, scorer, processor, score_cutoff, worst_score, optimal_score, scorer_kwargs)


cdef inline extract_dict_f64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    proc_query = move(RF_StringWrapper(conv_sequence(query)))
    proc_choices = preprocess_dict(choices, processor)

    cdef vector[DictMatchElem[double]] results = extract_dict_impl[double](
        scorer_kwargs, scorer_flags, scorer, proc_query, proc_choices,
        get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
        get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
    )

    # due to score_cutoff not always completely filled
    if limit > <int64_t>results.size():
        limit = <int64_t>results.size()

    if limit >= <int64_t>results.size():
        algorithm.sort(results.begin(), results.end(), ExtractComp(scorer_flags))
    else:
        algorithm.partial_sort(results.begin(), results.begin() + <ptrdiff_t>limit, results.end(), ExtractComp(scorer_flags))
        results.resize(limit)

    # copy elements into Python List
    result_list = PyList_New(<Py_ssize_t>limit)
    for i in range(limit):
        result_item = (<object>results[i].choice.obj, results[i].score, <object>results[i].key.obj)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list


cdef inline extract_dict_i64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    proc_query = move(RF_StringWrapper(conv_sequence(query)))
    proc_choices = preprocess_dict(choices, processor)

    cdef vector[DictMatchElem[int64_t]] results = extract_dict_impl[int64_t](
        scorer_kwargs, scorer_flags, scorer, proc_query, proc_choices,
        get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
        get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
    )

    # due to score_cutoff not always completely filled
    if limit > <int64_t>results.size():
        limit = <int64_t>results.size()

    if limit >= <int64_t>results.size():
        algorithm.sort(results.begin(), results.end(), ExtractComp(scorer_flags))
    else:
        algorithm.partial_sort(results.begin(), results.begin() + <ptrdiff_t>limit, results.end(), ExtractComp(scorer_flags))
        results.resize(limit)

    # copy elements into Python List
    result_list = PyList_New(<Py_ssize_t>limit)
    for i in range(limit):
        result_item = (<object>results[i].choice.obj, results[i].score, <object>results[i].key.obj)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list


cdef inline extract_dict_size_t(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    proc_query = move(RF_StringWrapper(conv_sequence(query)))
    proc_choices = preprocess_dict(choices, processor)

    cdef vector[DictMatchElem[size_t]] results = extract_dict_impl[size_t](
        scorer_kwargs, scorer_flags, scorer, proc_query, proc_choices,
        get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
        get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
    )

    # due to score_cutoff not always completely filled
    if limit > <int64_t>results.size():
        limit = <int64_t>results.size()

    if limit >= <int64_t>results.size():
        algorithm.sort(results.begin(), results.end(), ExtractComp(scorer_flags))
    else:
        algorithm.partial_sort(results.begin(), results.begin() + <ptrdiff_t>limit, results.end(), ExtractComp(scorer_flags))
        results.resize(limit)

    # copy elements into Python List
    result_list = PyList_New(<Py_ssize_t>limit)
    for i in range(limit):
        result_item = (<object>results[i].choice.obj, results[i].score, <object>results[i].key.obj)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list


cdef inline extract_dict(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    flags = scorer_flags.flags

    if flags & RF_SCORER_FLAG_RESULT_F64:
        return extract_dict_f64(
            query, choices, scorer, scorer_flags, processor, limit, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        return extract_dict_size_t(
            query, choices, scorer, scorer_flags, processor, limit, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        return extract_dict_i64(
            query, choices, scorer, scorer_flags, processor, limit, score_cutoff, score_hint, scorer_kwargs
        )

    raise ValueError("scorer does not properly use the C-API")


cdef inline extract_list_f64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    proc_query = move(RF_StringWrapper(conv_sequence(query)))
    proc_choices = preprocess_list(choices, processor)

    cdef vector[ListMatchElem[double]] results = extract_list_impl[double](
        scorer_kwargs, scorer_flags, scorer, proc_query, proc_choices,
        get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
        get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
    )

    # due to score_cutoff not always completely filled
    if limit > <int64_t>results.size():
        limit = <int64_t>results.size()

    if limit >= <int64_t>results.size():
        algorithm.sort(results.begin(), results.end(), ExtractComp(scorer_flags))
    else:
        algorithm.partial_sort(results.begin(), results.begin() + <ptrdiff_t>limit, results.end(), ExtractComp(scorer_flags))
        results.resize(limit)

    # copy elements into Python List
    result_list = PyList_New(<Py_ssize_t>limit)
    for i in range(limit):
        result_item = (<object>results[i].choice.obj, results[i].score, results[i].index)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list


cdef inline extract_list_i64(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    proc_query = move(RF_StringWrapper(conv_sequence(query)))
    proc_choices = preprocess_list(choices, processor)

    cdef vector[ListMatchElem[int64_t]] results = extract_list_impl[int64_t](
        scorer_kwargs, scorer_flags, scorer, proc_query, proc_choices,
        get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
        get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
    )

    # due to score_cutoff not always completely filled
    if limit > <int64_t>results.size():
        limit = <int64_t>results.size()

    if limit >= <int64_t>results.size():
        algorithm.sort(results.begin(), results.end(), ExtractComp(scorer_flags))
    else:
        algorithm.partial_sort(results.begin(), results.begin() + <ptrdiff_t>limit, results.end(), ExtractComp(scorer_flags))
        results.resize(limit)

    # copy elements into Python List
    result_list = PyList_New(<Py_ssize_t>limit)
    for i in range(limit):
        result_item = (<object>results[i].choice.obj, results[i].score, results[i].index)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list


cdef inline extract_list_size_t(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    proc_query = move(RF_StringWrapper(conv_sequence(query)))
    proc_choices = preprocess_list(choices, processor)

    cdef vector[ListMatchElem[size_t]] results = extract_list_impl[size_t](
        scorer_kwargs, scorer_flags, scorer, proc_query, proc_choices,
        get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
        get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
    )

    # due to score_cutoff not always completely filled
    if limit > <int64_t>results.size():
        limit = <int64_t>results.size()

    if limit >= <int64_t>results.size():
        algorithm.sort(results.begin(), results.end(), ExtractComp(scorer_flags))
    else:
        algorithm.partial_sort(results.begin(), results.begin() + <ptrdiff_t>limit, results.end(), ExtractComp(scorer_flags))
        results.resize(limit)

    # copy elements into Python List
    result_list = PyList_New(<Py_ssize_t>limit)
    for i in range(limit):
        result_item = (<object>results[i].choice.obj, results[i].score, results[i].index)
        Py_INCREF(result_item)
        PyList_SET_ITEM(result_list, <Py_ssize_t>i, result_item)

    return result_list


cdef inline extract_list(
    query, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    int64_t limit,
    score_cutoff,
    score_hint,
    const RF_Kwargs* scorer_kwargs
):
    flags = scorer_flags.flags

    if flags & RF_SCORER_FLAG_RESULT_F64:
        return extract_list_f64(
            query, choices, scorer, scorer_flags, processor, limit, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        return extract_list_size_t(
            query, choices, scorer, scorer_flags, processor, limit, score_cutoff, score_hint, scorer_kwargs
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        return extract_list_i64(
            query, choices, scorer, scorer_flags, processor, limit, score_cutoff, score_hint, scorer_kwargs
        )

    raise ValueError("scorer does not properly use the C-API")


cdef inline py_extract_dict(query, choices, scorer, processor, int64_t limit, double score_cutoff, worst_score, optimal_score, dict scorer_kwargs):
    cdef bool lowest_score_worst = optimal_score > worst_score
    cdef object score = None
    cdef list result_list = []

    for choice_key, choice in choices.items():
        if is_none(choice):
            continue

        if processor is not None:
            score = scorer(query, processor(choice), **scorer_kwargs)
        else:
            score = scorer(query, choice, **scorer_kwargs)

        if lowest_score_worst:
            if score >= score_cutoff:
                result_list.append((choice, score, choice_key))
        else:
            if score <= score_cutoff:
                result_list.append((choice, score, choice_key))

    if lowest_score_worst:
        return heapq.nlargest(limit, result_list, key=lambda i: i[1])
    else:
        return heapq.nsmallest(limit, result_list, key=lambda i: i[1])


cdef inline py_extract_list(query, choices, scorer, processor, int64_t limit, double score_cutoff, worst_score, optimal_score, dict scorer_kwargs):
    cdef bool lowest_score_worst = optimal_score > worst_score
    cdef object score = None
    cdef list result_list = []
    cdef int64_t i

    for i, choice in enumerate(choices):
        if is_none(choice):
            continue

        if processor is not None:
            score = scorer(query, processor(choice), **scorer_kwargs)
        else:
            score = scorer(query, choice, **scorer_kwargs)

        if lowest_score_worst:
            if score >= score_cutoff:
                result_list.append((choice, score, i))
        else:
            if score <= score_cutoff:
                result_list.append((choice, score, i))

    if lowest_score_worst:
        return heapq.nlargest(limit, result_list, key=lambda i: i[1])
    else:
        return heapq.nsmallest(limit, result_list, key=lambda i: i[1])


def extract(query, choices, *, scorer=WRatio, processor=None, limit=5, score_cutoff=None, score_hint=None, scorer_kwargs=None):
    cdef RF_Scorer* scorer_context = NULL
    cdef RF_ScorerFlags scorer_flags
    cdef int64_t c_limit
    cdef int64_t choices_len
    scorer_kwargs = scorer_kwargs.copy() if scorer_kwargs else {}

    setupPandas()

    if is_none(query):
        return []

    try:
        choices_len = <int64_t>len(choices)
    except TypeError:
        # handle generators. In Theory we could retrieve the length later on while
        # preprocessing the choices, but this is good enough for now
        choices = list(choices)
        choices_len = <int64_t>len(choices)

    c_limit = choices_len
    if limit is not None:
        c_limit = min(c_limit, <int64_t>limit)

    if c_limit == 1:
        res = extractOne(
            query,
            choices,
            processor=processor,
            scorer=scorer,
            score_cutoff=score_cutoff,
            score_hint=score_hint,
            scorer_kwargs=scorer_kwargs,
        )
        if res is None:
            return []
        return [res]

    # preprocess the query
    if callable(processor):
        query = processor(query)

    scorer_capsule = getattr(scorer, '_RF_Scorer', scorer)
    if PyCapsule_IsValid(scorer_capsule, NULL):
        scorer_context = <RF_Scorer*>PyCapsule_GetPointer(scorer_capsule, NULL)

    if scorer_context and scorer_context.version == SCORER_STRUCT_VERSION:
        kwargs_context = RF_KwargsWrapper()
        scorer_context.kwargs_init(&kwargs_context.kwargs, scorer_kwargs)
        scorer_context.get_scorer_flags(&kwargs_context.kwargs, &scorer_flags)

        if hasattr(choices, "items"):
            return extract_dict(query, choices, scorer_context, &scorer_flags,
                processor, c_limit, score_cutoff, score_hint, &kwargs_context.kwargs)
        else:
            return extract_list(query, choices, scorer_context, &scorer_flags,
                processor, c_limit, score_cutoff, score_hint, &kwargs_context.kwargs)


    worst_score, optimal_score = get_scorer_flags_py(scorer, scorer_kwargs)
    # the scorer has to be called through Python
    if score_cutoff is None:
        score_cutoff = worst_score

    scorer_kwargs["score_cutoff"] = score_cutoff

    if hasattr(choices, "items"):
        return py_extract_dict(query, choices, scorer, processor, c_limit, score_cutoff, worst_score, optimal_score, scorer_kwargs)
    else:
        return py_extract_list(query, choices, scorer, processor, c_limit, score_cutoff, worst_score, optimal_score, scorer_kwargs)


def extract_iter(query, choices, *, scorer=WRatio, processor=None, score_cutoff=None, score_hint=None, scorer_kwargs=None):
    cdef RF_Scorer* scorer_context = NULL
    cdef RF_ScorerFlags scorer_flags
    cdef RF_Preprocessor* processor_context = NULL
    cdef RF_KwargsWrapper kwargs_context

    def extract_iter_dict_f64():
        """
        implementation of extract_iter for dict, scorer using RapidFuzz C-API with the result type
        float64
        """
        cdef RF_String proc_str
        cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
        cdef double c_score_hint = get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
        query_proc = RF_StringWrapper(conv_sequence(query))

        cdef RF_ScorerFunc scorer_func
        scorer_context.scorer_func_init(
            &scorer_func, &kwargs_context.kwargs, 1, &query_proc.string
        )
        cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)
        cdef bool lowest_score_worst = is_lowest_score_worst[double](&scorer_flags)
        cdef double score

        for choice_key, choice in choices.items():
            if is_none(choice):
                continue

            # use RapidFuzz C-Api
            if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
                processor_context.preprocess(choice, &proc_str)
                choice_proc = RF_StringWrapper(proc_str)
            elif processor is not None:
                proc_choice = processor(choice)
                if is_none(proc_choice):
                    continue

                choice_proc = RF_StringWrapper(conv_sequence(proc_choice))
            else:
                choice_proc = RF_StringWrapper(conv_sequence(choice))

            ScorerFunc.call(&choice_proc.string, c_score_cutoff, c_score_hint, &score)
            if lowest_score_worst:
                if score >= c_score_cutoff:
                    yield (choice, score, choice_key)
            else:
                if score <= c_score_cutoff:
                    yield (choice, score, choice_key)

    def extract_iter_dict_i64():
        """
        implementation of extract_iter for dict, scorer using RapidFuzz C-API with the result type
        int64_t
        """
        cdef RF_String proc_str
        cdef int64_t c_score_cutoff = get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
        cdef int64_t c_score_hint = get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
        query_proc = RF_StringWrapper(conv_sequence(query))

        cdef RF_ScorerFunc scorer_func
        scorer_context.scorer_func_init(
            &scorer_func, &kwargs_context.kwargs, 1, &query_proc.string
        )
        cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)
        cdef bool lowest_score_worst = is_lowest_score_worst[int64_t](&scorer_flags)
        cdef int64_t score

        for choice_key, choice in choices.items():
            if is_none(choice):
                continue

            # use RapidFuzz C-Api
            if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
                processor_context.preprocess(choice, &proc_str)
                choice_proc = RF_StringWrapper(proc_str)
            elif processor is not None:
                proc_choice = processor(choice)
                if is_none(proc_choice):
                    continue

                choice_proc = RF_StringWrapper(conv_sequence(proc_choice))
            else:
                choice_proc = RF_StringWrapper(conv_sequence(choice))

            ScorerFunc.call(&choice_proc.string, c_score_cutoff, c_score_hint, &score)
            if lowest_score_worst:
                if score >= c_score_cutoff:
                    yield (choice, score, choice_key)
            else:
                if score <= c_score_cutoff:
                    yield (choice, score, choice_key)

    def extract_iter_dict_size_t():
        """
        implementation of extract_iter for dict, scorer using RapidFuzz C-API with the result type
        size_t
        """
        cdef RF_String proc_str
        cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
        cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
        query_proc = RF_StringWrapper(conv_sequence(query))

        cdef RF_ScorerFunc scorer_func
        scorer_context.scorer_func_init(
            &scorer_func, &kwargs_context.kwargs, 1, &query_proc.string
        )
        cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)
        cdef bool lowest_score_worst = is_lowest_score_worst[size_t](&scorer_flags)
        cdef size_t score

        for choice_key, choice in choices.items():
            if is_none(choice):
                continue

            # use RapidFuzz C-Api
            if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
                processor_context.preprocess(choice, &proc_str)
                choice_proc = RF_StringWrapper(proc_str)
            elif processor is not None:
                proc_choice = processor(choice)
                if is_none(proc_choice):
                    continue

                choice_proc = RF_StringWrapper(conv_sequence(proc_choice))
            else:
                choice_proc = RF_StringWrapper(conv_sequence(choice))

            ScorerFunc.call(&choice_proc.string, c_score_cutoff, c_score_hint, &score)
            if lowest_score_worst:
                if score >= c_score_cutoff:
                    yield (choice, score, choice_key)
            else:
                if score <= c_score_cutoff:
                    yield (choice, score, choice_key)

    def extract_iter_list_f64():
        """
        implementation of extract_iter for list, scorer using RapidFuzz C-API with the result type
        float64
        """
        cdef RF_String proc_str
        cdef double c_score_cutoff = get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
        cdef double c_score_hint = get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64)
        query_proc = RF_StringWrapper(conv_sequence(query))

        cdef RF_ScorerFunc scorer_func
        scorer_context.scorer_func_init(
            &scorer_func, &kwargs_context.kwargs, 1, &query_proc.string
        )
        cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)
        cdef bool lowest_score_worst = is_lowest_score_worst[double](&scorer_flags)
        cdef double score

        for i, choice in enumerate(choices):
            if is_none(choice):
                continue

            # use RapidFuzz C-Api
            if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
                processor_context.preprocess(choice, &proc_str)
                choice_proc = RF_StringWrapper(proc_str)
            elif processor is not None:
                proc_choice = processor(choice)
                if is_none(proc_choice):
                    continue

                choice_proc = RF_StringWrapper(conv_sequence(proc_choice))
            else:
                choice_proc = RF_StringWrapper(conv_sequence(choice))

            ScorerFunc.call(&choice_proc.string, c_score_cutoff, c_score_hint, &score)
            if lowest_score_worst:
                if score >= c_score_cutoff:
                    yield (choice, score, i)
            else:
                if score <= c_score_cutoff:
                    yield (choice, score, i)

    def extract_iter_list_i64():
        """
        implementation of extract_iter for list, scorer using RapidFuzz C-API with the result type
        int64_t
        """
        cdef RF_String proc_str
        cdef int64_t c_score_cutoff = get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
        cdef int64_t c_score_hint = get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64)
        query_proc = RF_StringWrapper(conv_sequence(query))

        cdef RF_ScorerFunc scorer_func
        scorer_context.scorer_func_init(
            &scorer_func, &kwargs_context.kwargs, 1, &query_proc.string
        )
        cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)
        cdef bool lowest_score_worst = is_lowest_score_worst[int64_t](&scorer_flags)
        cdef int64_t score

        for i, choice in enumerate(choices):
            if is_none(choice):
                continue

            # use RapidFuzz C-Api
            if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
                processor_context.preprocess(choice, &proc_str)
                choice_proc = RF_StringWrapper(proc_str)
            elif processor is not None:
                proc_choice = processor(choice)
                if is_none(proc_choice):
                    continue

                choice_proc = RF_StringWrapper(conv_sequence(proc_choice))
            else:
                choice_proc = RF_StringWrapper(conv_sequence(choice))

            ScorerFunc.call(&choice_proc.string, c_score_cutoff, c_score_hint, &score)
            if lowest_score_worst:
                if score >= c_score_cutoff:
                    yield (choice, score, i)
            else:
                if score <= c_score_cutoff:
                    yield (choice, score, i)

    def extract_iter_list_size_t():
        """
        implementation of extract_iter for list, scorer using RapidFuzz C-API with the result type
        size_t
        """
        cdef RF_String proc_str
        cdef size_t c_score_cutoff = get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
        cdef size_t c_score_hint = get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet)
        query_proc = RF_StringWrapper(conv_sequence(query))

        cdef RF_ScorerFunc scorer_func
        scorer_context.scorer_func_init(
            &scorer_func, &kwargs_context.kwargs, 1, &query_proc.string
        )
        cdef RF_ScorerWrapper ScorerFunc = RF_ScorerWrapper(scorer_func)
        cdef bool lowest_score_worst = is_lowest_score_worst[size_t](&scorer_flags)
        cdef size_t score

        for i, choice in enumerate(choices):
            if is_none(choice):
                continue

            # use RapidFuzz C-Api
            if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
                processor_context.preprocess(choice, &proc_str)
                choice_proc = RF_StringWrapper(proc_str)
            elif processor is not None:
                proc_choice = processor(choice)
                if is_none(proc_choice):
                    continue

                choice_proc = RF_StringWrapper(conv_sequence(proc_choice))
            else:
                choice_proc = RF_StringWrapper(conv_sequence(choice))

            ScorerFunc.call(&choice_proc.string, c_score_cutoff, c_score_hint, &score)
            if lowest_score_worst:
                if score >= c_score_cutoff:
                    yield (choice, score, i)
            else:
                if score <= c_score_cutoff:
                    yield (choice, score, i)

    def py_extract_iter_dict(worst_score, optimal_score):
        """
        implementation of extract_iter for:
          - type of choices = dict
          - scorer = python function
        """
        cdef bool lowest_score_worst = optimal_score > worst_score

        for choice_key, choice in choices.items():
            if is_none(choice):
                continue

            if processor is not None:
                score = scorer(query, processor(choice), **scorer_kwargs)
            else:
                score = scorer(query, choice, **scorer_kwargs)

            if lowest_score_worst:
                if score >= score_cutoff:
                    yield (choice, score, choice_key)
            else:
                if score <= score_cutoff:
                    yield (choice, score, choice_key)

    def py_extract_iter_list(worst_score, optimal_score):
        """
        implementation of extract_iter for:
          - type of choices = list
          - scorer = python function
        """
        cdef bool lowest_score_worst = optimal_score > worst_score
        cdef int64_t i

        for i, choice in enumerate(choices):
            if is_none(choice):
                continue

            if processor is not None:
                score = scorer(query, processor(choice), **scorer_kwargs)
            else:
                score = scorer(query, choice, **scorer_kwargs)

            if lowest_score_worst:
                if score >= score_cutoff:
                    yield (choice, score, i)
            else:
                if score <= score_cutoff:
                    yield (choice, score, i)

    scorer_kwargs = scorer_kwargs.copy() if scorer_kwargs else {}

    setupPandas()

    if is_none(query):
        # finish generator
        return

    # preprocess the query
    if callable(processor):
        query = processor(query)

    scorer_capsule = getattr(scorer, '_RF_Scorer', scorer)
    if PyCapsule_IsValid(scorer_capsule, NULL):
        scorer_context = <RF_Scorer*>PyCapsule_GetPointer(scorer_capsule, NULL)

    if scorer_context and scorer_context.version == SCORER_STRUCT_VERSION:
        kwargs_context = RF_KwargsWrapper()
        scorer_context.kwargs_init(&kwargs_context.kwargs, scorer_kwargs)
        scorer_context.get_scorer_flags(&kwargs_context.kwargs, &scorer_flags)

        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

        if hasattr(choices, "items"):
            if scorer_flags.flags & RF_SCORER_FLAG_RESULT_F64:
                yield from extract_iter_dict_f64()
                return
            elif scorer_flags.flags & RF_SCORER_FLAG_RESULT_SIZE_T:
                yield from extract_iter_dict_size_t()
                return
            elif scorer_flags.flags & RF_SCORER_FLAG_RESULT_I64:
                yield from extract_iter_dict_i64()
                return
        else:
            if scorer_flags.flags & RF_SCORER_FLAG_RESULT_F64:
                yield from extract_iter_list_f64()
                return
            elif scorer_flags.flags & RF_SCORER_FLAG_RESULT_SIZE_T:
                yield from extract_iter_list_size_t()
                return
            elif scorer_flags.flags & RF_SCORER_FLAG_RESULT_I64:
                yield from extract_iter_list_i64()
                return

        raise ValueError("scorer does not properly use the C-API")

    worst_score, optimal_score = get_scorer_flags_py(scorer, scorer_kwargs)
    # the scorer has to be called through Python
    if score_cutoff is None:
        score_cutoff = worst_score

    scorer_kwargs["score_cutoff"] = score_cutoff

    if hasattr(choices, "items"):
        yield from py_extract_iter_dict(worst_score, optimal_score)
    else:
        yield from py_extract_iter_list(worst_score, optimal_score)


FLOAT32 = MatrixType.FLOAT32
FLOAT64 = MatrixType.FLOAT64
INT8 = MatrixType.INT8
INT16 = MatrixType.INT16
INT32 = MatrixType.INT32
INT64 = MatrixType.INT64
UINT8 = MatrixType.UINT8
UINT16 = MatrixType.UINT16
UINT32 = MatrixType.UINT32
UINT64 = MatrixType.UINT64

cdef inline vector[PyObjectWrapper] preprocess_py(queries, processor) except *:
    cdef vector[PyObjectWrapper] proc_queries
    cdef int64_t queries_len = <int64_t>len(queries)
    proc_queries.reserve(queries_len)

    # processor None/False
    if not processor:
        for query in queries:
            proc_queries.emplace_back(<PyObject*>query)
    # processor has to be called through python
    else:
        for query in queries:
            proc_query = query if is_none(query) else processor(query)
            proc_queries.emplace_back(<PyObject*>proc_query)

    return move(proc_queries)

cdef inline vector[RF_StringWrapper] preprocess(const RF_ScorerFlags* scorer_flags, queries, processor) except *:
    cdef vector[RF_StringWrapper] proc_queries
    cdef int64_t queries_len = <int64_t>len(queries)
    cdef RF_String proc_str
    cdef RF_Preprocessor* processor_context = NULL
    flags = scorer_flags.flags
    proc_queries.reserve(queries_len)

    # No processor
    if not processor:
        for query in queries:
            conv_sequence_with_none(query, &proc_str)
            if proc_str.data == NULL:
                if flags & RF_SCORER_NONE_IS_WORST_SCORE:
                    proc_queries.emplace_back()
                else:
                    raise ValueError(f"passed unsupported element {query}")
            else:
                proc_queries.emplace_back(proc_str, <PyObject*>query)
    else:
        processor_capsule = getattr(processor, '_RF_Preprocess', processor)
        if PyCapsule_IsValid(processor_capsule, NULL):
            processor_context = <RF_Preprocessor*>PyCapsule_GetPointer(processor_capsule, NULL)

        # use RapidFuzz C-Api
        if processor_context != NULL and processor_context.version == PREPROCESSOR_STRUCT_VERSION:
            for query in queries:
                if is_none(query) and flags & RF_SCORER_NONE_IS_WORST_SCORE:
                    proc_queries.emplace_back()
                else:
                    processor_context.preprocess(query, &proc_str)
                    proc_queries.emplace_back(proc_str, <PyObject*>query)

        # Call Processor through Python
        else:
            for query in queries:
                if is_none(query) and flags & RF_SCORER_NONE_IS_WORST_SCORE:
                    proc_queries.emplace_back()
                else:
                    proc_query = processor(query)
                    proc_queries.emplace_back(conv_sequence(proc_query), <PyObject*>proc_query)

    return move(proc_queries)

cdef inline MatrixType dtype_to_type_num_f64(dtype) except MatrixType.UNDEFINED:
    if dtype is None:
        return MatrixType.FLOAT32
    return <MatrixType>dtype

cdef inline MatrixType dtype_to_type_num_i64(dtype) except MatrixType.UNDEFINED:
    if dtype is None:
        return MatrixType.INT32
    return <MatrixType>dtype

cdef inline MatrixType dtype_to_type_num_size_t(dtype) except MatrixType.UNDEFINED:
    if dtype is None:
        return MatrixType.UINT32
    return <MatrixType>dtype

cdef inline MatrixType dtype_to_type_num_py(dtype, scorer, dict scorer_kwargs) except MatrixType.UNDEFINED:
    import numpy as np

    if dtype is not None:
        return <MatrixType>dtype

    params = getattr(scorer, "_RF_ScorerPy", None)
    if params is not None:
        flags = params["get_scorer_flags"](**scorer_kwargs)
        if <int>flags["flags"] & RF_SCORER_FLAG_RESULT_I64:
            return MatrixType.INT32
        if <int>flags["flags"] & RF_SCORER_FLAG_RESULT_SIZE_T:
            return MatrixType.UINT32
        return MatrixType.FLOAT32

    return MatrixType.FLOAT32


from cpython cimport Py_buffer
from libcpp.vector cimport vector


cdef class Matrix:

    cdef RfMatrix matrix
    cdef Py_ssize_t shape[2]
    cdef Py_ssize_t strides[2]
    cdef bint vector_output

    def __cinit__(self, bint vector_output=False):
        self.vector_output = vector_output

    def __getbuffer__(self, Py_buffer *buffer, int flags):

        if self.vector_output:
            self.shape[0] = self.matrix.m_rows
            self.strides[0] = self.matrix.get_dtype_size()
            buffer.ndim = 1
        else:
            self.shape[0] = self.matrix.m_rows
            self.shape[1] = self.matrix.m_cols
            self.strides[1] = self.matrix.get_dtype_size()
            self.strides[0] = self.matrix.m_cols * self.strides[1]
            buffer.ndim = 2

        buffer.buf = <char *>self.matrix.m_matrix
        buffer.format = <char *>self.matrix.get_format()
        buffer.internal = NULL
        buffer.itemsize = self.matrix.get_dtype_size()
        buffer.len = self.matrix.m_rows * self.matrix.m_cols * self.matrix.get_dtype_size()
        buffer.obj = self
        buffer.readonly = 0
        buffer.shape = self.shape
        buffer.strides = self.strides
        buffer.suboffsets = NULL

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

cdef Matrix cpdist_cpp(
    queries, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    score_multiplier,
    dtype,
    int c_workers,
    const RF_Kwargs* scorer_kwargs
):
    proc_queries = preprocess(scorer_flags, queries, processor)
    proc_choices = preprocess(scorer_flags, choices, processor)
    flags = scorer_flags.flags
    cdef Matrix matrix = Matrix(vector_output=True)

    if flags & RF_SCORER_FLAG_RESULT_F64:
        matrix.matrix = cpdist_cpp_impl[double](
            scorer_kwargs, scorer, proc_queries, proc_choices,
            dtype_to_type_num_f64(dtype),
            c_workers,
            get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
            get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
            <double>score_multiplier,
            scorer_flags.worst_score.f64
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        matrix.matrix = cpdist_cpp_impl[size_t](
            scorer_kwargs, scorer, proc_queries, proc_choices,
            dtype_to_type_num_size_t(dtype),
            c_workers,
            get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
            get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
            <size_t>score_multiplier,
            scorer_flags.worst_score.sizet
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        matrix.matrix = cpdist_cpp_impl[int64_t](
            scorer_kwargs, scorer, proc_queries, proc_choices,
            dtype_to_type_num_i64(dtype),
            c_workers,
            get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
            get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
            <int64_t>score_multiplier,
            scorer_flags.worst_score.i64
        )
    else:
        raise ValueError("scorer does not properly use the C-API")

    return matrix

cdef Matrix cdist_two_lists(
    queries, choices,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    score_multiplier,
    dtype,
    int c_workers,
    const RF_Kwargs* scorer_kwargs
):
    proc_queries = preprocess(scorer_flags, queries, processor)
    proc_choices = preprocess(scorer_flags, choices, processor)
    flags = scorer_flags.flags
    cdef Matrix matrix = Matrix()

    if flags & RF_SCORER_FLAG_RESULT_F64:
        matrix.matrix = cdist_two_lists_impl[double](
            scorer_flags,
            scorer_kwargs, scorer, proc_queries, proc_choices,
            dtype_to_type_num_f64(dtype),
            c_workers,
            get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
            get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
            <double>score_multiplier,
            scorer_flags.worst_score.f64,
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        matrix.matrix = cdist_two_lists_impl[size_t](
            scorer_flags,
            scorer_kwargs, scorer, proc_queries, proc_choices,
            dtype_to_type_num_size_t(dtype),
            c_workers,
            get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
            get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
            <size_t>score_multiplier,
            scorer_flags.worst_score.sizet
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        matrix.matrix = cdist_two_lists_impl[int64_t](
            scorer_flags,
            scorer_kwargs, scorer, proc_queries, proc_choices,
            dtype_to_type_num_i64(dtype),
            c_workers,
            get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
            get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
            <int64_t>score_multiplier,
            scorer_flags.worst_score.i64
        )
    else:
        raise ValueError("scorer does not properly use the C-API")

    return matrix

cdef Matrix cdist_single_list(
    queries,
    RF_Scorer* scorer,
    const RF_ScorerFlags* scorer_flags,
    processor,
    score_cutoff,
    score_hint,
    score_multiplier,
    dtype,
    int c_workers,
    const RF_Kwargs* scorer_kwargs
):
    proc_queries = preprocess(scorer_flags, queries, processor)
    flags = scorer_flags.flags
    cdef Matrix matrix = Matrix()

    if flags & RF_SCORER_FLAG_RESULT_F64:
        matrix.matrix = cdist_single_list_impl[double](
            scorer_flags,
            scorer_kwargs, scorer, proc_queries,
            dtype_to_type_num_f64(dtype),
            c_workers,
            get_score_cutoff_f64(score_cutoff, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
            get_score_cutoff_f64(score_hint, scorer_flags.worst_score.f64, scorer_flags.optimal_score.f64),
            <double>score_multiplier,
            scorer_flags.worst_score.f64
        )
    elif flags & RF_SCORER_FLAG_RESULT_SIZE_T:
        matrix.matrix = cdist_single_list_impl[size_t](
            scorer_flags,
            scorer_kwargs, scorer, proc_queries,
            dtype_to_type_num_size_t(dtype),
            c_workers,
            get_score_cutoff_size_t(score_cutoff, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
            get_score_cutoff_size_t(score_hint, scorer_flags.worst_score.sizet, scorer_flags.optimal_score.sizet),
            <size_t>score_multiplier,
            scorer_flags.worst_score.sizet
        )
    elif flags & RF_SCORER_FLAG_RESULT_I64:
        matrix.matrix = cdist_single_list_impl[int64_t](
            scorer_flags,
            scorer_kwargs, scorer, proc_queries,
            dtype_to_type_num_i64(dtype),
            c_workers,
            get_score_cutoff_i64(score_cutoff, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
            get_score_cutoff_i64(score_hint, scorer_flags.worst_score.i64, scorer_flags.optimal_score.i64),
            <int64_t>score_multiplier,
            scorer_flags.worst_score.i64
        )
    else:
        raise ValueError("scorer does not properly use the C-API")

    return matrix


@cython.boundscheck(False)
@cython.wraparound(False)
cdef cdist_py(queries, choices, scorer, processor, score_cutoff, score_multiplier, dtype, workers, dict scorer_kwargs):
    # todo this should handle two similar sequences more efficiently

    proc_queries = preprocess_py(queries, processor)
    proc_choices = preprocess_py(choices, processor)
    cdef double score
    cdef Matrix matrix = Matrix()
    c_dtype = dtype_to_type_num_py(dtype, scorer, scorer_kwargs)
    matrix.matrix = RfMatrix(c_dtype, proc_queries.size(), proc_choices.size())

    scorer_kwargs["score_cutoff"] = score_cutoff

    for i in range(proc_queries.size()):
        for j in range(proc_choices.size()):
            score = scorer(<object>proc_queries[i].obj, <object>proc_choices[j].obj, **scorer_kwargs)
            matrix.matrix.set(i, j, score * <double>score_multiplier)

    return matrix

def cdist(queries, choices, *, scorer=ratio, processor=None, score_cutoff=None, score_hint=None, score_multiplier=1, dtype=None, workers=1, scorer_kwargs=None):
    cdef RF_Scorer* scorer_context = NULL
    cdef RF_ScorerFlags scorer_flags
    cdef bool is_orig_scorer

    setupPandas()

    scorer_kwargs = scorer_kwargs.copy() if scorer_kwargs else {}

    scorer_capsule = getattr(scorer, '_RF_Scorer', scorer)
    if PyCapsule_IsValid(scorer_capsule, NULL):
        scorer_context = <RF_Scorer*>PyCapsule_GetPointer(scorer_capsule, NULL)

    is_orig_scorer = getattr(scorer, '_RF_OriginalScorer', None) is scorer

    if is_orig_scorer and scorer_context and scorer_context.version == SCORER_STRUCT_VERSION:
        kwargs_context = RF_KwargsWrapper()
        scorer_context.kwargs_init(&kwargs_context.kwargs, scorer_kwargs)
        scorer_context.get_scorer_flags(&kwargs_context.kwargs, &scorer_flags)

        # scorer(queries[i], choices[j]) == scorer(queries[j], choices[i])
        if scorer_flags.flags & RF_SCORER_FLAG_SYMMETRIC and queries is choices:
            return cdist_single_list(
                queries, scorer_context, &scorer_flags, processor,
                score_cutoff, score_hint, score_multiplier,
                dtype, workers, &kwargs_context.kwargs)
        else:
            return cdist_two_lists(
                queries, choices, scorer_context, &scorer_flags, processor,
                score_cutoff, score_hint, score_multiplier,
                dtype, workers, &kwargs_context.kwargs)

    return cdist_py(queries, choices, scorer, processor, score_cutoff, score_multiplier, dtype, workers, scorer_kwargs)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef cpdist_py(queries, choices, scorer, processor, score_cutoff, score_multiplier, dtype, workers, dict scorer_kwargs):

    proc_queries = preprocess_py(queries, processor)
    proc_choices = preprocess_py(choices, processor)
    cdef double score
    cdef Matrix matrix = Matrix(vector_output=True)
    c_dtype = dtype_to_type_num_py(dtype, scorer, scorer_kwargs)
    matrix.matrix = RfMatrix(c_dtype, proc_queries.size(), 1)

    scorer_kwargs["score_cutoff"] = score_cutoff

    for i in range(proc_queries.size()):
        score = scorer(<object>proc_queries[i].obj, <object>proc_choices[i].obj, **scorer_kwargs)
        matrix.matrix.set(i, 0, score * <double>score_multiplier)

    return matrix

def cpdist(queries, choices, *, scorer=ratio, processor=None, score_cutoff=None, score_hint=None, score_multiplier=1, dtype=None, workers=1, scorer_kwargs=None):

    if len(queries) != len(choices):
        error_message = "Length of queries and choices must be the same!"
        raise ValueError(error_message)

    cdef RF_Scorer* scorer_context = NULL
    cdef RF_ScorerFlags scorer_flags
    cdef bool is_orig_scorer

    setupPandas()

    scorer_kwargs = scorer_kwargs.copy() if scorer_kwargs else {}

    scorer_capsule = getattr(scorer, '_RF_Scorer', scorer)
    if PyCapsule_IsValid(scorer_capsule, NULL):
        scorer_context = <RF_Scorer*>PyCapsule_GetPointer(scorer_capsule, NULL)

    is_orig_scorer = getattr(scorer, '_RF_OriginalScorer', None) is scorer

    if is_orig_scorer and scorer_context and scorer_context.version == SCORER_STRUCT_VERSION:
        kwargs_context = RF_KwargsWrapper()
        scorer_context.kwargs_init(&kwargs_context.kwargs, scorer_kwargs)
        scorer_context.get_scorer_flags(&kwargs_context.kwargs, &scorer_flags)

        return cpdist_cpp(
            queries, choices, scorer_context, &scorer_flags, processor,
            score_cutoff, score_hint, score_multiplier,
            dtype, workers, &kwargs_context.kwargs)

    return cpdist_py(queries, choices, scorer, processor, score_cutoff, score_multiplier, dtype, workers, scorer_kwargs)



from __future__ import annotations

from collections.abc import Collection, Generator, Hashable, Iterable, Mapping, Sequence
from typing import (
    Any,
    Callable,
    Protocol,
    TypeVar,
    overload,
)

from rapidfuzz.fuzz import WRatio, ratio

_StringType = Sequence[Hashable]
_StringType1 = TypeVar("_StringType1", bound=Sequence[Hashable])
_StringType2 = TypeVar("_StringType2", bound=Sequence[Hashable])
_UnprocessedType1 = TypeVar("_UnprocessedType1")
_UnprocessedType2 = TypeVar("_UnprocessedType2")
_KeyType = TypeVar("_KeyType")
_ResultType = TypeVar("_ResultType", int, float)

_StringType1_contra = TypeVar("_StringType1_contra", contravariant=True, bound=Sequence[Hashable])
_StringType2_contra = TypeVar("_StringType2_contra", contravariant=True, bound=Sequence[Hashable])
_ResultType_contra = TypeVar("_ResultType_contra", int, float, contravariant=True)
_ResultType_co = TypeVar("_ResultType_co", int, float, covariant=True)

class _Scorer(Protocol[_StringType1_contra, _StringType2_contra, _ResultType_contra, _ResultType_co]):
    def __call__(
        self, __s1: _StringType1_contra, __s2: _StringType2_contra, *, score_cutoff: _ResultType_contra | None
    ) -> _ResultType_co: ...

# mypy wants defaults to be valid for every possible parameterization of a generic function
# so add separate overloads for the default version
@overload
def extractOne(
    query: Sequence[Hashable] | None,
    choices: Mapping[_KeyType, _StringType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: None = None,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_StringType2, float, _KeyType]: ...
@overload
def extractOne(
    query: Sequence[Hashable] | None,
    choices: Iterable[_StringType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: None = None,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_StringType2, float, int]: ...
@overload
def extractOne(
    query: _UnprocessedType1 | None,
    choices: Mapping[_KeyType, _UnprocessedType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_UnprocessedType2, float, _KeyType]: ...
@overload
def extractOne(
    query: _UnprocessedType1 | None,
    choices: Iterable[_UnprocessedType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_UnprocessedType2, float, int]: ...
@overload
def extractOne(
    query: _StringType1 | None,
    choices: Mapping[_KeyType, _StringType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
    processor: None = None,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_StringType2, _ResultType, _KeyType]: ...
@overload
def extractOne(
    query: _StringType1 | None,
    choices: Iterable[_StringType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
    processor: None = None,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_StringType2, _ResultType, int]: ...
@overload
def extractOne(
    query: _UnprocessedType1 | None,
    choices: Mapping[_KeyType, _UnprocessedType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_UnprocessedType2, _ResultType, _KeyType]: ...
@overload
def extractOne(
    query: _UnprocessedType1 | None,
    choices: Iterable[_UnprocessedType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> tuple[_UnprocessedType2, _ResultType, int]: ...

# mypy wants defaults to be valid for every possible parameterization of a generic function
# so add separate overloads for the default version
@overload
def extract(
    query: Sequence[Hashable] | None,
    choices: Mapping[_KeyType, _StringType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: None = None,
    limit: int | None = 5,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_StringType2, float, _KeyType]]: ...
@overload
def extract(
    query: Sequence[Hashable] | None,
    choices: Iterable[_StringType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: None = None,
    limit: int | None = 5,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_StringType2, float, int]]: ...
@overload
def extract(
    query: _UnprocessedType1 | None,
    choices: Mapping[_KeyType, _UnprocessedType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
    limit: int | None = 5,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_UnprocessedType2, float, _KeyType]]: ...
@overload
def extract(
    query: _UnprocessedType1 | None,
    choices: Iterable[_UnprocessedType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
    limit: int | None = 5,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_UnprocessedType2, float, int]]: ...
@overload
def extract(
    query: _StringType1 | None,
    choices: Mapping[_KeyType, _StringType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
    processor: None = None,
    limit: int | None = 5,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_StringType2, _ResultType, _KeyType]]: ...
@overload
def extract(
    query: _StringType1 | None,
    choices: Collection[_StringType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
    processor: None = None,
    limit: int | None = 5,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_StringType2, _ResultType, int]]: ...
@overload
def extract(
    query: _UnprocessedType1 | None,
    choices: Mapping[_KeyType, _UnprocessedType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
    limit: int | None = 5,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_UnprocessedType2, _ResultType, _KeyType]]: ...
@overload
def extract(
    query: _UnprocessedType1 | None,
    choices: Collection[_UnprocessedType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
    limit: int | None = 5,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> list[tuple[_UnprocessedType2, _ResultType, int]]: ...

# mypy wants defaults to be valid for every possible parameterization of a generic function
# so add separate overloads for the default version
@overload
def extract_iter(
    query: Sequence[Hashable] | None,
    choices: Mapping[_KeyType, _StringType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: None = None,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_StringType2, float, _KeyType], None, None]: ...
@overload
def extract_iter(
    query: Sequence[Hashable] | None,
    choices: Iterable[_StringType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: None = None,
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_StringType2, float, int], None, None]: ...
@overload
def extract_iter(
    query: _UnprocessedType1 | None,
    choices: Mapping[_KeyType, _UnprocessedType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_UnprocessedType2, float, _KeyType], None, None]: ...
@overload
def extract_iter(
    query: _UnprocessedType1 | None,
    choices: Iterable[_UnprocessedType2 | None],
    *,
    scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = WRatio,
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
    score_cutoff: float | None = None,
    score_hint: float | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_UnprocessedType2, float, int], None, None]: ...
@overload
def extract_iter(
    query: _StringType1 | None,
    choices: Mapping[_KeyType, _StringType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
    processor: None = None,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_StringType2, _ResultType, _KeyType], None, None]: ...
@overload
def extract_iter(
    query: _StringType1 | None,
    choices: Iterable[_StringType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
    processor: None = None,
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_StringType2, _ResultType, int], None, None]: ...
@overload
def extract_iter(
    query: _UnprocessedType1 | None,
    choices: Mapping[_KeyType, _UnprocessedType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_UnprocessedType2, _ResultType, _KeyType], None, None]: ...
@overload
def extract_iter(
    query: _UnprocessedType1 | None,
    choices: Iterable[_UnprocessedType2 | None],
    *,
    scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
    processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
    score_cutoff: _ResultType | None = None,
    score_hint: _ResultType | None = None,
    scorer_kwargs: dict[str, Any] | None = None,
) -> Generator[tuple[_UnprocessedType2, _ResultType, int], None, None]: ...

try:
    import numpy as np

    @overload
    def cdist(
        queries: Iterable[Sequence[Hashable] | None],
        choices: Iterable[Sequence[Hashable] | None],
        *,
        scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = ratio,
        processor: None = None,
        score_cutoff: float | None = None,
        score_hint: float | None = None,
        score_multiplier: float = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cdist(
        queries: Iterable[_UnprocessedType1 | None],
        choices: Iterable[_UnprocessedType2 | None],
        *,
        scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = ratio,
        processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
        score_cutoff: float | None = None,
        score_hint: float | None = None,
        score_multiplier: float = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cdist(
        queries: Iterable[_StringType1 | None],
        choices: Iterable[_StringType2 | None],
        *,
        scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
        processor: None = None,
        score_cutoff: _ResultType | None = None,
        score_hint: _ResultType | None = None,
        score_multiplier: _ResultType = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cdist(
        queries: Iterable[_UnprocessedType1 | None],
        choices: Iterable[_UnprocessedType2 | None],
        *,
        scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
        processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
        score_cutoff: _ResultType | None = None,
        score_hint: _ResultType | None = None,
        score_multiplier: _ResultType = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cpdist(
        queries: Iterable[Sequence[Hashable] | None],
        choices: Iterable[Sequence[Hashable] | None],
        *,
        scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = ratio,
        processor: None = None,
        score_cutoff: float | None = None,
        score_hint: float | None = None,
        score_multiplier: float = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cpdist(
        queries: Iterable[_UnprocessedType1 | None],
        choices: Iterable[_UnprocessedType2 | None],
        *,
        scorer: _Scorer[Sequence[Hashable], Sequence[Hashable], float, float] = ratio,
        processor: Callable[[_UnprocessedType1 | _UnprocessedType2], Sequence[Hashable]],
        score_cutoff: float | None = None,
        score_hint: float | None = None,
        score_multiplier: float = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cpdist(
        queries: Iterable[_StringType1 | None],
        choices: Iterable[_StringType2 | None],
        *,
        scorer: _Scorer[_StringType1, _StringType2, _ResultType, _ResultType],
        processor: None = None,
        score_cutoff: _ResultType | None = None,
        score_hint: _ResultType | None = None,
        score_multiplier: _ResultType = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...
    @overload
    def cpdist(
        queries: Iterable[_UnprocessedType1 | None],
        choices: Iterable[_UnprocessedType2 | None],
        *,
        scorer: _Scorer[_StringType1, _StringType1, _ResultType, _ResultType],
        processor: Callable[[_UnprocessedType1 | _UnprocessedType2], _StringType1],
        score_cutoff: _ResultType | None = None,
        score_hint: _ResultType | None = None,
        score_multiplier: _ResultType = 1,
        dtype: np.dtype | None = None,
        workers: int = 1,
        scorer_kwargs: dict[str, Any] | None = None,
    ) -> np.ndarray: ...

except ImportError:
    pass



from __future__ import annotations

from collections.abc import Generator, Hashable, Sequence

import numpy as np
import numpy.typing as npt

from rapidfuzz import fuzz, process
from rapidfuzz.distance import Levenshtein


def string_preprocessor(a: str) -> str:
    return a


class MyClass:
    def __init__(self) -> None:
        self.a: str = ""


def valid_scorer_seq1(s1: Sequence[Hashable], s2: Sequence[Hashable], score_cutoff: float | None = None) -> float:
    return 1.0


def valid_scorer_str1(s1: str, s2: str, score_cutoff: float | None = None) -> float:
    return 1.0


def valid_scorer_str2(s1: str, s2: str, *, score_cutoff: float | None) -> float:
    return 1.0


def valid_scorer_str3(string1: str, string2: str, *, score_cutoff: float | None) -> float:
    return 1.0


def invalid_scorer_str1(s1: str, s2: str, cutoff: float | None = None) -> float:
    return 1.0


def invalid_scorer_str2(s1: str, s2: str) -> float:
    return 1.0


def invalid_scorer_str3(s1: str, s2: str, score_cutoff: int | None = None) -> float:
    return 1.0


def invalid_scorer_str4(s1: int, s2: str, score_cutoff: float | None = None) -> float:
    return 1.0


def invalid_scorer_str5(s1: str, s2: int, score_cutoff: float | None = None) -> float:
    return 1.0


def test_extractOne():
    _a: tuple[str, float, int] = process.extractOne("", [""])
    _a = process.extractOne("", [""], scorer=fuzz.ratio)
    _b: tuple[str, int, int] = process.extractOne("", [""], scorer=fuzz.ratio)  # type: ignore [assignment]
    _b = process.extractOne("", [""])  # type: ignore [assignment]
    _b = process.extractOne("", [""], scorer=Levenshtein.distance)
    _a = process.extractOne(1, [""], scorer=fuzz.ratio)  # type: ignore [call-overload]
    _a = process.extractOne("", [1], scorer=fuzz.ratio)  # type: ignore [call-overload]
    _a = process.extractOne(list(""), [""], scorer=fuzz.ratio)
    _c: tuple[list[str], float, int] = process.extractOne("", [list("")], scorer=fuzz.ratio)
    _a = process.extractOne("", [""], processor=string_preprocessor)
    _a = process.extractOne(list(""), [""], processor=string_preprocessor)  # type: ignore [arg-type]
    _c = process.extractOne("", [list("")], processor=string_preprocessor)  # type: ignore [arg-type]
    _d: tuple[MyClass, float, int] = process.extractOne(MyClass(), [MyClass()], processor=lambda x: x.a)
    _a = process.extractOne("", [""], scorer=valid_scorer_str1)
    _a = process.extractOne("", [""], scorer=valid_scorer_str2)
    _a = process.extractOne("", [""], scorer=valid_scorer_str3)
    _a = process.extractOne("", [""], scorer=invalid_scorer_str1)  # type: ignore [call-overload]
    _a = process.extractOne("", [""], scorer=invalid_scorer_str2)  # type: ignore [call-overload]
    _a = process.extractOne("", [""], scorer=invalid_scorer_str3)  # type: ignore [call-overload]
    _a = process.extractOne("", [""], scorer=invalid_scorer_str4)  # type: ignore [call-overload]
    _a = process.extractOne("", [""], scorer=invalid_scorer_str5)  # type: ignore [call-overload]
    _a = process.extractOne(list(""), [""], scorer=valid_scorer_str1)  # type: ignore [call-overload]
    _c = process.extractOne("", [list("")], scorer=valid_scorer_str1)  # type: ignore [call-overload]
    _a = process.extractOne(list(""), [""], scorer=valid_scorer_seq1)
    _c = process.extractOne("", [list("")], scorer=valid_scorer_seq1)


def test_extract():
    _a: list[tuple[str, float, int]] = process.extract("", [""])
    _a = process.extract("", [""], scorer=fuzz.ratio)
    _b: list[tuple[str, int, int]] = process.extract("", [""], scorer=fuzz.ratio)  # type: ignore [assignment]
    _b = process.extract("", [""])  # type: ignore [assignment]
    _b = process.extract("", [""], scorer=Levenshtein.distance)
    _a = process.extract(1, [""], scorer=fuzz.ratio)  # type: ignore [call-overload]
    _a = process.extract("", [1], scorer=fuzz.ratio)  # type: ignore [call-overload]
    _a = process.extract(list(""), [""], scorer=fuzz.ratio)
    _c: list[tuple[list[str], float, int]] = process.extract("", [list("")], scorer=fuzz.ratio)
    _a = process.extract("", [""], processor=string_preprocessor)
    _a = process.extract(list(""), [""], processor=string_preprocessor)  # type: ignore [arg-type]
    _c = process.extract("", [list("")], processor=string_preprocessor)  # type: ignore [arg-type]
    _d: list[tuple[MyClass, float, int]] = process.extract(MyClass(), [MyClass()], processor=lambda x: x.a)
    _a = process.extract("", [""], scorer=valid_scorer_str1)
    _a = process.extract("", [""], scorer=valid_scorer_str2)
    _a = process.extract("", [""], scorer=valid_scorer_str3)
    _a = process.extract("", [""], scorer=invalid_scorer_str1)  # type: ignore [call-overload]
    _a = process.extract("", [""], scorer=invalid_scorer_str2)  # type: ignore [call-overload]
    _a = process.extract("", [""], scorer=invalid_scorer_str3)  # type: ignore [call-overload]
    _a = process.extract("", [""], scorer=invalid_scorer_str4)  # type: ignore [call-overload]
    _a = process.extract("", [""], scorer=invalid_scorer_str5)  # type: ignore [call-overload]
    _a = process.extract(list(""), [""], scorer=valid_scorer_str1)  # type: ignore [call-overload]
    _c = process.extract("", [list("")], scorer=valid_scorer_str1)  # type: ignore [call-overload]
    _a = process.extract(list(""), [""], scorer=valid_scorer_seq1)
    _c = process.extract("", [list("")], scorer=valid_scorer_seq1)


def test_extract_iter():
    _a: Generator[tuple[str, float, int], None, None] = process.extract_iter("", [""])
    _a = process.extract_iter("", [""], scorer=fuzz.ratio)
    _b: Generator[tuple[str, int, int], None, None] = process.extract_iter("", [""], scorer=fuzz.ratio)  # type: ignore [assignment]
    _b = process.extract_iter("", [""])  # type: ignore [assignment]
    _b = process.extract_iter("", [""], scorer=Levenshtein.distance)
    _a = process.extract_iter(1, [""], scorer=fuzz.ratio)  # type: ignore [call-overload]
    _a = process.extract_iter("", [1], scorer=fuzz.ratio)  # type: ignore [call-overload]
    _a = process.extract_iter(list(""), [""], scorer=fuzz.ratio)
    _c: Generator[tuple[list[str], float, int], None, None] = process.extract_iter("", [list("")], scorer=fuzz.ratio)
    _a = process.extract_iter("", [""], processor=string_preprocessor)
    _a = process.extract_iter(list(""), [""], processor=string_preprocessor)  # type: ignore [arg-type]
    _c = process.extract_iter("", [list("")], processor=string_preprocessor)  # type: ignore [arg-type]
    _d: Generator[tuple[MyClass, float, int], None, None] = process.extract_iter(
        MyClass(), [MyClass()], processor=lambda x: x.a
    )
    _a = process.extract_iter("", [""], scorer=valid_scorer_str1)
    _a = process.extract_iter("", [""], scorer=valid_scorer_str2)
    _a = process.extract_iter("", [""], scorer=valid_scorer_str3)
    _a = process.extract_iter("", [""], scorer=invalid_scorer_str1)  # type: ignore [call-overload]
    _a = process.extract_iter("", [""], scorer=invalid_scorer_str2)  # type: ignore [call-overload]
    _a = process.extract_iter("", [""], scorer=invalid_scorer_str3)  # type: ignore [call-overload]
    _a = process.extract_iter("", [""], scorer=invalid_scorer_str4)  # type: ignore [call-overload]
    _a = process.extract_iter("", [""], scorer=invalid_scorer_str5)  # type: ignore [call-overload]
    _a = process.extract_iter(list(""), [""], scorer=valid_scorer_str1)  # type: ignore [call-overload]
    _c = process.extract_iter("", [list("")], scorer=valid_scorer_str1)  # type: ignore [call-overload]
    _a = process.extract_iter(list(""), [""], scorer=valid_scorer_seq1)
    _c = process.extract_iter("", [list("")], scorer=valid_scorer_seq1)


def test_cdist():
    _a: npt.NDArray[np.float32] = process.cdist([""], [""])
    _a = process.cdist([""], [""], scorer=fuzz.ratio)
    # todo these should be be rejected, but I didn't find a way to properly type the return type so far
    # _b: npt.NDArray[np.int32] = process.cdist([""], [""], scorer=fuzz.ratio)  type: ignore [assignment]
    # _b = process.cdist([""], [""])  type: ignore [assignment]
    _b: npt.NDArray[np.int32] = process.cdist([""], [""], scorer=Levenshtein.distance)
    _a = process.cdist([1], [""], scorer=fuzz.ratio)  # type: ignore [list-item]
    _a = process.cdist([""], [1], scorer=fuzz.ratio)  # type: ignore [list-item]
    _a = process.cdist([list("")], [""], scorer=fuzz.ratio)
    _a = process.cdist([""], [""], processor=string_preprocessor)
    _a = process.cdist([list("")], [""], processor=string_preprocessor)  # type: ignore [arg-type]
    _a = process.cdist([""], [list("")], processor=string_preprocessor)  # type: ignore [arg-type]
    _a = process.cdist([MyClass()], [MyClass()], processor=lambda x: x.a)
    _a = process.cdist([""], [""], scorer=valid_scorer_str1)
    _a = process.cdist([""], [""], scorer=valid_scorer_str2)
    _a = process.cdist([""], [""], scorer=valid_scorer_str3)
    _a = process.cdist([""], [""], scorer=invalid_scorer_str1)  # type: ignore [arg-type]
    _a = process.cdist([""], [""], scorer=invalid_scorer_str2)  # type: ignore [arg-type]
    _a = process.cdist([""], [""], scorer=invalid_scorer_str3)  # type: ignore [arg-type]
    _a = process.cdist([""], [""], scorer=invalid_scorer_str4)  # type: ignore [arg-type]
    _a = process.cdist([""], [""], scorer=invalid_scorer_str5)  # type: ignore [arg-type]
    _a = process.cdist([list("")], [""], scorer=valid_scorer_str1)  # type: ignore [arg-type]
    _a = process.cdist([""], [list("")], scorer=valid_scorer_str1)  # type: ignore [arg-type]
    _a = process.cdist([list("")], [""], scorer=valid_scorer_seq1)
    _a = process.cdist([""], [list("")], scorer=valid_scorer_seq1)


def test_cpdist():
    _a: npt.NDArray[np.float32] = process.cpdist([""], [""])
    _a = process.cpdist([""], [""], scorer=fuzz.ratio)
    # todo these should be be rejected, but I didn't find a way to properly type the return type so far
    # _b: npt.NDArray[np.int32] = process.cdist([""], [""], scorer=fuzz.ratio)  type: ignore [assignment]
    # _b = process.cdist([""], [""])  type: ignore [assignment]
    _b: npt.NDArray[np.int32] = process.cpdist([""], [""], scorer=Levenshtein.distance)
    _a = process.cpdist([1], [""], scorer=fuzz.ratio)  # type: ignore [list-item]
    _a = process.cpdist([""], [1], scorer=fuzz.ratio)  # type: ignore [list-item]
    _a = process.cpdist([list("")], [""], scorer=fuzz.ratio)
    _a = process.cpdist([""], [""], processor=string_preprocessor)
    _a = process.cpdist([list("")], [""], processor=string_preprocessor)  # type: ignore [arg-type]
    _a = process.cpdist([""], [list("")], processor=string_preprocessor)  # type: ignore [arg-type]
    _a = process.cpdist([MyClass()], [MyClass()], processor=lambda x: x.a)
    _a = process.cpdist([""], [""], scorer=valid_scorer_str1)
    _a = process.cpdist([""], [""], scorer=valid_scorer_str2)
    _a = process.cpdist([""], [""], scorer=valid_scorer_str3)
    _a = process.cpdist([""], [""], scorer=invalid_scorer_str1)  # type: ignore [arg-type]
    _a = process.cpdist([""], [""], scorer=invalid_scorer_str2)  # type: ignore [arg-type]
    _a = process.cpdist([""], [""], scorer=invalid_scorer_str3)  # type: ignore [arg-type]
    _a = process.cpdist([""], [""], scorer=invalid_scorer_str4)  # type: ignore [arg-type]
    _a = process.cpdist([""], [""], scorer=invalid_scorer_str5)  # type: ignore [arg-type]
    _a = process.cpdist([list("")], [""], scorer=valid_scorer_str1)  # type: ignore [arg-type]
    _a = process.cpdist([""], [list("")], scorer=valid_scorer_str1)  # type: ignore [arg-type]
    _a = process.cpdist([list("")], [""], scorer=valid_scorer_seq1)
    _a = process.cpdist([""], [list("")], scorer=valid_scorer_seq1)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
# Copyright (C) 2025 Max Bachmann
from __future__ import annotations

import subprocess
from pathlib import Path

import isort

format = """
# SPDX-License-Identifier: MIT
# Copyright (C) 2025 Max Bachmann
# This file is generated by tools/generate_python.py
from __future__ import annotations

import contextlib
import os

from rapidfuzz._feature_detector import AVX2, SSE2, supports

__all__ = [
    {exports}
]

_impl = os.environ.get("RAPIDFUZZ_IMPLEMENTATION")
if _impl == "cpp":
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from {module}_cpp_avx2 import {includes} # pyright: ignore[reportMissingImports]
            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from {module}_cpp_sse2 import {includes} # pyright: ignore[reportMissingImports]
            imported = True

    if not imported:
        from {module}_cpp import {includes} # pyright: ignore[reportMissingImports]
elif _impl == "python":
    from {module}_py import {includes}
else:
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from {module}_cpp_avx2 import {includes} # pyright: ignore[reportMissingImports]
            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from {module}_cpp_sse2 import {includes} # pyright: ignore[reportMissingImports]
            imported = True

    if not imported:
        with contextlib.suppress(ImportError):
            from {module}_cpp import {includes} # pyright: ignore[reportMissingImports]
            imported = True

    if not imported:
        from {module}_py import {includes}
"""


def generate(module, importModule, includes):
    segments = module.split(".")
    segments[-1] = segments[-1] + ".py"
    path = Path(__file__).parent.parent / "src"
    for segment in segments:
        path = path.joinpath(segment)

    print(f"generating {path}")
    if isinstance(includes, list):
        includesStr = ", ".join(includes)
        exportStr = ", ".join(f'"{x}"' for x in includes)
    else:
        includesStr = ", ".join(f"{k} as {v}" for k, v in includes.items())
        exportStr = ", ".join(f'"{x}"' for x in includes.values())

    formatted = format.format(module=importModule, includes=includesStr, exports=exportStr)

    config = isort.settings.Config(combine_as_imports=True)
    formatted = isort.code(formatted, config=config)

    with path.open("w", encoding="utf-8") as f:
        f.write(formatted)

    print(f"formatting {path} using ruff")
    subprocess.run(["ruff", "format", path], check=False)
    subprocess.run(["ruff", "check", path, "--fix"], check=False)


generate(
    "rapidfuzz.fuzz",
    "rapidfuzz.fuzz",
    [
        "ratio",
        "partial_ratio",
        "partial_ratio_alignment",
        "token_sort_ratio",
        "token_set_ratio",
        "token_ratio",
        "partial_token_sort_ratio",
        "partial_token_set_ratio",
        "partial_token_ratio",
        "WRatio",
        "QRatio",
    ],
)


generate(
    "rapidfuzz.process",
    "rapidfuzz.process",
    [
        "extract",
        "extractOne",
        "extract_iter",
        "cdist",
        "cpdist",
    ],
)

generate(
    "rapidfuzz.utils",
    "rapidfuzz.utils",
    [
        "default_process",
    ],
)

generate(
    "rapidfuzz.distance._initialize",
    "rapidfuzz.distance._initialize",
    [
        "Editop",
        "Editops",
        "Opcode",
        "Opcodes",
        "ScoreAlignment",
        "MatchingBlock",
    ],
)

generate(
    "rapidfuzz.distance.DamerauLevenshtein",
    "rapidfuzz.distance.metrics",
    {
        "damerau_levenshtein_distance": "distance",
        "damerau_levenshtein_similarity": "similarity",
        "damerau_levenshtein_normalized_distance": "normalized_distance",
        "damerau_levenshtein_normalized_similarity": "normalized_similarity",
    },
)

generate(
    "rapidfuzz.distance.Hamming",
    "rapidfuzz.distance.metrics",
    {
        "hamming_distance": "distance",
        "hamming_similarity": "similarity",
        "hamming_normalized_distance": "normalized_distance",
        "hamming_normalized_similarity": "normalized_similarity",
        "hamming_editops": "editops",
        "hamming_opcodes": "opcodes",
    },
)

generate(
    "rapidfuzz.distance.Indel",
    "rapidfuzz.distance.metrics",
    {
        "indel_distance": "distance",
        "indel_similarity": "similarity",
        "indel_normalized_distance": "normalized_distance",
        "indel_normalized_similarity": "normalized_similarity",
        "indel_editops": "editops",
        "indel_opcodes": "opcodes",
    },
)

generate(
    "rapidfuzz.distance.Jaro",
    "rapidfuzz.distance.metrics",
    {
        "jaro_distance": "distance",
        "jaro_similarity": "similarity",
        "jaro_normalized_distance": "normalized_distance",
        "jaro_normalized_similarity": "normalized_similarity",
    },
)

generate(
    "rapidfuzz.distance.JaroWinkler",
    "rapidfuzz.distance.metrics",
    {
        "jaro_winkler_distance": "distance",
        "jaro_winkler_similarity": "similarity",
        "jaro_winkler_normalized_distance": "normalized_distance",
        "jaro_winkler_normalized_similarity": "normalized_similarity",
    },
)

generate(
    "rapidfuzz.distance.LCSseq",
    "rapidfuzz.distance.metrics",
    {
        "lcs_seq_distance": "distance",
        "lcs_seq_similarity": "similarity",
        "lcs_seq_normalized_distance": "normalized_distance",
        "lcs_seq_normalized_similarity": "normalized_similarity",
        "lcs_seq_editops": "editops",
        "lcs_seq_opcodes": "opcodes",
    },
)

generate(
    "rapidfuzz.distance.Levenshtein",
    "rapidfuzz.distance.metrics",
    {
        "levenshtein_distance": "distance",
        "levenshtein_similarity": "similarity",
        "levenshtein_normalized_distance": "normalized_distance",
        "levenshtein_normalized_similarity": "normalized_similarity",
        "levenshtein_editops": "editops",
        "levenshtein_opcodes": "opcodes",
    },
)

generate(
    "rapidfuzz.distance.OSA",
    "rapidfuzz.distance.metrics",
    {
        "osa_distance": "distance",
        "osa_similarity": "similarity",
        "osa_normalized_distance": "normalized_distance",
        "osa_normalized_similarity": "normalized_similarity",
    },
)

generate(
    "rapidfuzz.distance.Postfix",
    "rapidfuzz.distance.metrics",
    {
        "postfix_distance": "distance",
        "postfix_similarity": "similarity",
        "postfix_normalized_distance": "normalized_distance",
        "postfix_normalized_similarity": "normalized_similarity",
    },
)

generate(
    "rapidfuzz.distance.Prefix",
    "rapidfuzz.distance.metrics",
    {
        "prefix_distance": "distance",
        "prefix_similarity": "similarity",
        "prefix_normalized_distance": "normalized_distance",
        "prefix_normalized_similarity": "normalized_similarity",
    },
)



from __future__ import annotations

import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory

scorer_test_boilerplate = """
from __future__ import annotations

from rapidfuzz.distance import *  # pyright: ignore [reportWildcardImportFromLibrary]
from rapidfuzz import fuzz

def string_preprocessor(a: str) -> str:
    return a

class MyClass:
    def __init__(self) -> None:
        self.a: str = ""

"""

scorer_test_int = (
    scorer_test_boilerplate
    + """
def test():
    _a: int = {scorer}("", "")
    _a = {scorer}("", 1)  # type: ignore [call-overload]
    _a = {scorer}(1, "")  # type: ignore [call-overload]
    _a = {scorer}(list(""), "")
    _a = {scorer}("", list(""))
    _a = {scorer}("", "", score_cutoff=0.5)  # type: ignore [call-overload]
    _a = {scorer}("", "", processor=string_preprocessor)
    _a = {scorer}("", list(""), processor=string_preprocessor)  # type: ignore [arg-type]
    _a = {scorer}(list(""), "", processor=string_preprocessor)  # type: ignore [arg-type]
    _a = {scorer}(MyClass(), MyClass(), processor=lambda x: x.a)
"""
)

scorer_test_float = (
    scorer_test_boilerplate
    + """
def test():
    _a: float = {scorer}("", "")
    _b: int = {scorer}("", "")  # type: ignore [assignment]
    _a = {scorer}("", 1)  # type: ignore [call-overload]
    _a = {scorer}(1, "")  # type: ignore [call-overload]
    _a = {scorer}(list(""), "")
    _a = {scorer}("", list(""))
    _a = {scorer}("", "", processor=string_preprocessor)
    _a = {scorer}("", list(""), processor=string_preprocessor)  # type: ignore [arg-type]
    _a = {scorer}(list(""), "", processor=string_preprocessor)  # type: ignore [arg-type]
    _a = {scorer}(MyClass(), MyClass(), processor=lambda x: x.a)
"""
)


def test_scorer(scorer, res_type):
    with TemporaryDirectory() as d:
        f_name = Path(d) / "scorer_test.py"
        with open(f_name, "w") as f:
            if res_type is int:
                f.write(scorer_test_int.format(scorer=scorer))
            else:
                f.write(scorer_test_float.format(scorer=scorer))

        print(f"testing {scorer}")
        subprocess.run(["python", "-m", "mypy", str(f_name), "--warn-unused-ignores"], check=True)
        subprocess.run(
            [
                "python",
                "-m",
                "pyright",
                "-p",
                str(Path(__file__).resolve().parent / "pyrightconfig.json"),
                str(f_name),
            ],
            check=True,
        )


for module in ("DamerauLevenshtein", "Hamming", "Indel", "LCSseq", "Levenshtein", "OSA", "Postfix", "Prefix"):
    test_scorer(f"{module}.distance", int)
    test_scorer(f"{module}.similarity", int)
    test_scorer(f"{module}.normalized_distance", float)
    test_scorer(f"{module}.normalized_similarity", float)

for module in ("Jaro", "JaroWinkler"):
    test_scorer(f"{module}.distance", float)
    test_scorer(f"{module}.similarity", float)
    test_scorer(f"{module}.normalized_distance", float)
    test_scorer(f"{module}.normalized_similarity", float)

test_scorer("fuzz.ratio", float)
test_scorer("fuzz.partial_ratio", float)
test_scorer("fuzz.token_ratio", float)
test_scorer("fuzz.token_set_ratio", float)
test_scorer("fuzz.token_sort_ratio", float)
test_scorer("fuzz.partial_token_ratio", float)
test_scorer("fuzz.partial_token_set_ratio", float)
test_scorer("fuzz.partial_token_sort_ratio", float)
test_scorer("fuzz.WRatio", float)
test_scorer("fuzz.QRatio", float)



from __future__ import annotations

from array import array

import pytest

from rapidfuzz import fuzz_cpp, fuzz_py, utils_cpp, utils_py
from rapidfuzz.distance import ScoreAlignment

from .common import symmetric_scorer_tester


class fuzz:
    @staticmethod
    def ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def partial_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.partial_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.partial_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def partial_ratio_alignment(*args, **kwargs):
        dist1 = fuzz_cpp.partial_ratio_alignment(*args, **kwargs)
        dist2 = fuzz_py.partial_ratio_alignment(*args, **kwargs)
        if dist1 is None or dist2 is None:
            assert dist1 == dist2
        else:
            assert pytest.approx(dist1[0]) == dist2[0]
            assert list(dist1)[1:] == list(dist2)[1:]
        return dist1

    @staticmethod
    def token_sort_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.token_sort_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.token_sort_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def token_set_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.token_set_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.token_set_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def token_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.token_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.token_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def partial_token_sort_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.partial_token_sort_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.partial_token_sort_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def partial_token_set_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.partial_token_set_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.partial_token_set_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def partial_token_ratio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.partial_token_ratio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.partial_token_ratio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def WRatio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.WRatio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.WRatio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1

    @staticmethod
    def QRatio(*args, **kwargs):
        dist1 = symmetric_scorer_tester(fuzz_cpp.QRatio, *args, **kwargs)
        dist2 = symmetric_scorer_tester(fuzz_py.QRatio, *args, **kwargs)
        assert pytest.approx(dist1) == dist2
        return dist1


scorers = [
    fuzz.ratio,
    fuzz.partial_ratio,
    fuzz.token_sort_ratio,
    fuzz.token_set_ratio,
    fuzz.token_ratio,
    fuzz.partial_token_sort_ratio,
    fuzz.partial_token_set_ratio,
    fuzz.partial_token_ratio,
    fuzz.WRatio,
    fuzz.QRatio,
]


def test_no_processor():
    assert fuzz.ratio("new york mets", "new york mets") == 100
    assert fuzz.ratio("new york mets", "new YORK mets") != 100


def test_partial_ratio():
    assert fuzz.partial_ratio("new york mets", "the wonderful new york mets") == 100


def test_token_sort_ratio():
    assert fuzz.token_sort_ratio("new york mets", "new york mets") == 100


def testPartialTokenSortRatio():
    assert fuzz.partial_token_sort_ratio("new york mets", "new york mets") == 100
    assert fuzz.partial_token_sort_ratio("new york mets vs atlanta braves", "atlanta braves vs new york mets") == 100


def testTokenSetRatio():
    assert fuzz.token_set_ratio("new york mets vs atlanta braves", "atlanta braves vs new york mets") == 100
    assert fuzz.token_set_ratio("js", "vue js") == 100


def testPartialTokenSetRatio():
    assert fuzz.partial_token_set_ratio("new york mets vs atlanta braves", "atlanta braves vs new york mets") == 100


def testQuickRatioEqual():
    assert fuzz.QRatio("new york mets", "new york mets", processor=utils_cpp.default_process) == 100
    assert fuzz.QRatio("new york mets", "new york mets", processor=utils_py.default_process) == 100


def testQuickRatioCaseInsensitive():
    assert fuzz.QRatio("new york mets", "new YORK mets", processor=utils_cpp.default_process) == 100
    assert fuzz.QRatio("new york mets", "new YORK mets", processor=utils_py.default_process) == 100


def testQuickRatioNotEqual():
    assert fuzz.QRatio("new york mets", "the wonderful new york mets") != 100


def testWRatioEqual():
    assert fuzz.WRatio("new york mets", "new york mets", processor=utils_cpp.default_process) == 100
    assert fuzz.WRatio("new york mets", "new york mets", processor=utils_py.default_process) == 100


def testWRatioCaseInsensitive():
    assert fuzz.WRatio("new york mets", "new YORK mets", processor=utils_cpp.default_process) == 100
    assert fuzz.WRatio("new york mets", "new YORK mets", processor=utils_py.default_process) == 100


def testWRatioPartialMatch():
    # a partial match is scaled by .9
    assert fuzz.WRatio("new york mets", "the wonderful new york mets") == 90


def testWRatioMisorderedMatch():
    # misordered full matches are scaled by .95
    assert fuzz.WRatio("new york mets vs atlanta braves", "atlanta braves vs new york mets") == 95


def testWRatioUnicode():
    assert fuzz.WRatio("new york mets", "new york mets") == 100


def testQRatioUnicode():
    assert fuzz.WRatio("new york mets", "new york mets") == 100


def test_issue76():
    assert pytest.approx(fuzz.partial_ratio("physics 2 vid", "study physics physics 2")) == 81.81818
    assert fuzz.partial_ratio("physics 2 vid", "study physics physics 2 video") == 100


def test_issue90():
    assert pytest.approx(fuzz.partial_ratio("ax b", "a b a c b")) == 85.71428


def test_issue138():
    str1 = "a" * 65
    str2 = "a" + chr(256) + "a" * 63
    assert pytest.approx(fuzz.partial_ratio(str1, str2)) == 99.22481


def test_partial_ratio_alignment():
    a = "a certain string"
    s = "certain"
    assert fuzz.partial_ratio_alignment(s, a) == ScoreAlignment(100, 0, len(s), 2, 2 + len(s))
    assert fuzz.partial_ratio_alignment(a, s) == ScoreAlignment(100, 2, 2 + len(s), 0, len(s))
    assert fuzz.partial_ratio_alignment(None, "test") is None
    assert fuzz.partial_ratio_alignment("test", None) is None

    assert fuzz.partial_ratio_alignment("test", "tesx", score_cutoff=90) is None


def test_issue196():
    """
    fuzz.WRatio did not work correctly with score_cutoffs
    """
    assert pytest.approx(fuzz.WRatio("South Korea", "North Korea")) == 81.81818
    assert fuzz.WRatio("South Korea", "North Korea", score_cutoff=85.4) == 0.0
    assert fuzz.WRatio("South Korea", "North Korea", score_cutoff=85.5) == 0.0


def test_issue231():
    str1 = "er merkantilismus förderte handle und verkehr mit teils marktkonformen, teils dirigistischen maßnahmen."
    str2 = "ils marktkonformen, teils dirigistischen maßnahmen. an der schwelle zum 19. jahrhundert entstand ein neu"

    alignment = fuzz.partial_ratio_alignment(str1, str2)
    assert alignment.src_start == 0
    assert alignment.src_end == 103
    assert alignment.dest_start == 0
    assert alignment.dest_end == 51


def test_empty_string():
    """
    when both strings are empty this is either a perfect match or no match
    See https://github.com/rapidfuzz/RapidFuzz/issues/110
    """
    # perfect match
    assert fuzz.ratio("", "") == 100
    assert fuzz.partial_ratio("", "") == 100
    assert fuzz.token_sort_ratio("", "") == 100
    assert fuzz.partial_token_sort_ratio("", "") == 100
    assert fuzz.token_ratio("", "") == 100
    assert fuzz.partial_token_ratio("", "") == 100

    # no match
    assert fuzz.WRatio("", "") == 0
    assert fuzz.QRatio("", "") == 0
    assert fuzz.token_set_ratio("", "") == 0
    assert fuzz.partial_token_set_ratio("", "") == 0

    # perfect match when no words
    assert fuzz.token_set_ratio("    ", "    ") == 0
    assert fuzz.partial_token_set_ratio("    ", "    ") == 0


@pytest.mark.parametrize("scorer", scorers)
def test_invalid_input(scorer):
    """
    when invalid types are passed to a scorer an exception should be thrown
    """
    with pytest.raises(TypeError):
        scorer(1, 1, catch_exceptions=True)


@pytest.mark.parametrize("scorer", scorers)
def test_array(scorer):
    """
    arrays should be supported and treated in a compatible way to strings
    """
    # todo add support in pure python implementation
    assert scorer(
        array("u", "the wonderful new york mets"),
        array("u", "the wonderful new york mets"),
    )
    assert scorer("the wonderful new york mets", array("u", "the wonderful new york mets"))
    assert scorer(array("u", "the wonderful new york mets"), "the wonderful new york mets")


@pytest.mark.parametrize("scorer", scorers)
def test_bytes(scorer):
    """
    bytes should be supported and treated in a compatible way to strings
    """
    assert scorer(b"the wonderful new york mets", b"the wonderful new york mets") == 100
    assert scorer("the wonderful new york mets", b"the wonderful new york mets") == 100
    assert scorer(b"the wonderful new york mets", "the wonderful new york mets") == 100


@pytest.mark.parametrize("scorer", scorers)
def test_none_string(scorer):
    """
    when None is passed to a scorer the result should always be 0
    """
    assert scorer("test", None) == 0
    assert scorer(None, "test") == 0


@pytest.mark.parametrize("scorer", scorers)
def test_nan_string(scorer):
    """
    when float("nan") is passed to a scorer the result should always be 0
    """
    assert scorer("test", float("nan")) == 0
    assert scorer(float("nan"), "test") == 0


@pytest.mark.parametrize("scorer", scorers)
def test_simple_unicode_tests(scorer):
    """
    some very simple tests using unicode with scorers
    to catch relatively obvious implementation errors
    """
    s1 = "ÁÄ"
    s2 = "ABCD"
    assert scorer(s1, s2) == 0
    assert scorer(s1, s1) == 100


@pytest.mark.parametrize(
    "processor",
    [
        utils_cpp.default_process,
        lambda s: utils_cpp.default_process(s),
        utils_py.default_process,
        lambda s: utils_py.default_process(s),
    ],
)
@pytest.mark.parametrize("scorer", scorers)
def test_scorer_case_insensitive(processor, scorer):
    """
    each scorer should be able to preprocess strings properly
    """
    assert scorer("new york mets", "new YORK mets", processor=processor) == 100


@pytest.mark.parametrize("processor", [None, lambda s: s])
def test_ratio_case_censitive(processor):
    assert fuzz.ratio("new york mets", "new YORK mets", processor=processor) != 100


@pytest.mark.parametrize("scorer", scorers)
def test_custom_processor(scorer):
    """
    Any scorer should accept any type as s1 and s2, as long as it is a string
    after preprocessing.
    """
    s1 = ["chicago cubs vs new york mets", "CitiField", "2011-05-11", "8pm"]
    s2 = ["chicago cubs vs new york mets", "CitiFields", "2012-05-11", "9pm"]
    s3 = ["different string", "CitiFields", "2012-05-11", "9pm"]
    assert scorer(s1, s2, processor=lambda event: event[0]) == 100
    assert scorer(s2, s3, processor=lambda event: event[0]) != 100


@pytest.mark.parametrize("scorer", scorers)
def testIssue206(scorer):
    """
    test correct behavior of score_cutoff
    """
    score1 = scorer("South Korea", "North Korea")
    score2 = scorer("South Korea", "North Korea", score_cutoff=score1 - 0.0001)
    assert score1 == score2


@pytest.mark.parametrize("scorer", scorers)
def test_help(scorer):
    """
    test that all help texts can be printed without throwing an exception,
    since they are implemented in C++ as well
    """
    help(scorer)


def testIssue257():
    s1 = "aaaaaaaaaaaaaaaaaaaaaaaabacaaaaaaaabaaabaaaaaaaababbbbbbbbbbabbcb"
    s2 = "aaaaaaaaaaaaaaaaaaaaaaaababaaaaaaaabaaabaaaaaaaababbbbbbbbbbabbcb"
    score = fuzz.partial_ratio(s1, s2)
    assert pytest.approx(score) == 98.46153846153847
    score = fuzz.partial_ratio(s2, s1)
    assert pytest.approx(score) == 98.46153846153847



import rapidfuzz
from rapidfuzz.distance import metrics_py
from rapidfuzz.distance import metrics_cpp
rapidfuzz.distance.Levenshtein.distance('test', 'teste')
metrics_py.levenshtein_distance('test', 'teste')
metrics_cpp.levenshtein_distance('test', 'teste')


# SPDX-License-Identifier: MIT
# Copyright (C) 2025 Max Bachmann
# This file is generated by tools/generate_python.py
from __future__ import annotations

import contextlib
import os

from rapidfuzz._feature_detector import AVX2, SSE2, supports

__all__ = ["Editop", "Editops", "MatchingBlock", "Opcode", "Opcodes", "ScoreAlignment"]

_impl = os.environ.get("RAPIDFUZZ_IMPLEMENTATION")
if _impl == "cpp":
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance._initialize_cpp_avx2 import (  # pyright: ignore[reportMissingImports]
                Editop,
                Editops,
                MatchingBlock,
                Opcode,
                Opcodes,
                ScoreAlignment,
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance._initialize_cpp_sse2 import (  # pyright: ignore[reportMissingImports]
                Editop,
                Editops,
                MatchingBlock,
                Opcode,
                Opcodes,
                ScoreAlignment,
            )

            imported = True

    if not imported:
        from rapidfuzz.distance._initialize_cpp import (  # pyright: ignore[reportMissingImports]
            Editop,
            Editops,
            MatchingBlock,
            Opcode,
            Opcodes,
            ScoreAlignment,
        )
elif _impl == "python":
    from rapidfuzz.distance._initialize_py import (
        Editop,
        Editops,
        MatchingBlock,
        Opcode,
        Opcodes,
        ScoreAlignment,
    )
else:
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance._initialize_cpp_avx2 import (  # pyright: ignore[reportMissingImports]
                Editop,
                Editops,
                MatchingBlock,
                Opcode,
                Opcodes,
                ScoreAlignment,
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance._initialize_cpp_sse2 import (  # pyright: ignore[reportMissingImports]
                Editop,
                Editops,
                MatchingBlock,
                Opcode,
                Opcodes,
                ScoreAlignment,
            )

            imported = True

    if not imported:
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance._initialize_cpp import (  # pyright: ignore[reportMissingImports]
                Editop,
                Editops,
                MatchingBlock,
                Opcode,
                Opcodes,
                ScoreAlignment,
            )

            imported = True

    if not imported:
        from rapidfuzz.distance._initialize_py import (
            Editop,
            Editops,
            MatchingBlock,
            Opcode,
            Opcodes,
            ScoreAlignment,
        )



# distutils: language=c++
# cython: language_level=3, binding=True, linetrace=True

from cpp_common cimport RfEditops, RfOpcodes
from libcpp cimport bool
from libcpp.vector cimport vector


cdef class Editops:
    cdef RfEditops editops

cdef class Opcodes:
    cdef RfOpcodes opcodes

cdef class ScoreAlignment:
    cdef public object score
    cdef public Py_ssize_t src_start
    cdef public Py_ssize_t src_end
    cdef public Py_ssize_t dest_start
    cdef public Py_ssize_t dest_end



from __future__ import annotations

from rapidfuzz import utils_cpp, utils_py
from tests.distance.common import Postfix


def test_basic():
    assert Postfix.distance("", "") == 0
    assert Postfix.distance("test", "test") == 0
    assert Postfix.distance("aaaa", "bbbb") == 4


def test_score_cutoff():
    """
    test whether score_cutoff works correctly
    """
    assert Postfix.distance("abcd", "eebcd") == 2
    assert Postfix.distance("abcd", "eebcd", score_cutoff=4) == 2
    assert Postfix.distance("abcd", "eebcd", score_cutoff=3) == 2
    assert Postfix.distance("abcd", "eebcd", score_cutoff=2) == 2
    assert Postfix.distance("abcd", "eebcd", score_cutoff=1) == 2
    assert Postfix.distance("abcd", "eebcd", score_cutoff=0) == 1


def testCaseInsensitive():
    assert Postfix.distance("new york mets", "new YORK mets", processor=utils_cpp.default_process) == 0
    assert Postfix.distance("new york mets", "new YORK mets", processor=utils_py.default_process) == 0



# SPDX-License-Identifier: MIT
# Copyright (C) 2025 Max Bachmann
# This file is generated by tools/generate_python.py
from __future__ import annotations

import contextlib
import os

from rapidfuzz._feature_detector import AVX2, SSE2, supports

__all__ = ["distance", "normalized_distance", "normalized_similarity", "similarity"]

_impl = os.environ.get("RAPIDFUZZ_IMPLEMENTATION")
if _impl == "cpp":
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance.metrics_cpp_avx2 import (  # pyright: ignore[reportMissingImports]
                jaro_distance as distance,
                jaro_normalized_distance as normalized_distance,
                jaro_normalized_similarity as normalized_similarity,
                jaro_similarity as similarity,
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance.metrics_cpp_sse2 import (  # pyright: ignore[reportMissingImports]
                jaro_distance as distance,
                jaro_normalized_distance as normalized_distance,
                jaro_normalized_similarity as normalized_similarity,
                jaro_similarity as similarity,
            )

            imported = True

    if not imported:
        from rapidfuzz.distance.metrics_cpp import (  # pyright: ignore[reportMissingImports]
            jaro_distance as distance,
            jaro_normalized_distance as normalized_distance,
            jaro_normalized_similarity as normalized_similarity,
            jaro_similarity as similarity,
        )
elif _impl == "python":
    from rapidfuzz.distance.metrics_py import (
        jaro_distance as distance,
        jaro_normalized_distance as normalized_distance,
        jaro_normalized_similarity as normalized_similarity,
        jaro_similarity as similarity,
    )
else:
    imported = False
    if supports(AVX2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance.metrics_cpp_avx2 import (  # pyright: ignore[reportMissingImports]
                jaro_distance as distance,
                jaro_normalized_distance as normalized_distance,
                jaro_normalized_similarity as normalized_similarity,
                jaro_similarity as similarity,
            )

            imported = True

    if not imported and supports(SSE2):
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance.metrics_cpp_sse2 import (  # pyright: ignore[reportMissingImports]
                jaro_distance as distance,
                jaro_normalized_distance as normalized_distance,
                jaro_normalized_similarity as normalized_similarity,
                jaro_similarity as similarity,
            )

            imported = True

    if not imported:
        with contextlib.suppress(ImportError):
            from rapidfuzz.distance.metrics_cpp import (  # pyright: ignore[reportMissingImports]
                jaro_distance as distance,
                jaro_normalized_distance as normalized_distance,
                jaro_normalized_similarity as normalized_similarity,
                jaro_similarity as similarity,
            )

            imported = True

    if not imported:
        from rapidfuzz.distance.metrics_py import (
            jaro_distance as distance,
            jaro_normalized_distance as normalized_distance,
            jaro_normalized_similarity as normalized_similarity,
            jaro_similarity as similarity,
        )
